import argparse
import os
from pathlib import Path
import sys
from enum import Enum
import numpy as np
import pandas as pd
import gzip
from PIL import Image

from mcfnmr import __version__
from mcfnmr.utils.loading import parseBrukerSpectrum, parseUCSF
from mcfnmr.utils.rasterspectrum import RasterSpectrum
from mcfnmr.utils.pointspectrum import PointSpectrum
from mcfnmr.utils.plotting import plotRasterSpectrum
from mcfnmr.config import csv_column_name_links


class InType(Enum):
    UCSF = 1
    BRUKER = 2
    BITMAP = 3
    PEAKLIST = 4


def saveLibAsCSV(lib, libname, libfile):    
    df = {"name": [], "weight": [], "1H": [], "13C": []}
    cpds = sorted(lib.keys())
    for c in cpds:
        spec = lib[c]
        df["name"].extend([c] * spec.size())
        df["weight"].extend(spec.weights)
        df["1H"].extend(spec.coords[:, 1])
        df["13C"].extend(spec.coords[:, 0])
    pd.DataFrame(df).to_csv(libfile, sep=",")
    print(f"\nWrote lib '{libname}' to '{libfile}'.\n")


def parse_args():
    arg_parser = argparse.ArgumentParser(
        prog="spec2csv",
        description=f"""
        spec2csv - part of MCFNMR (v{__version__}).
        Tool to convert a different types of spectrum files into a csv (or csv.gz) file usable as target for mcfnmr.
        Usage:  spec2csv --file=INPUT_FILE [--rangex=X1,X2 --rangey=Y1,Y2 --irange=I1,I2 --bindims=NX,NY]
        """,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    arg_parser.add_argument(
        "--inputfile",
        "-f",
        default=None,
        dest="infile",
        help="Spectrum file describing the intensity distribution. Currently supported: "
        "ucsf-files, bruker-txt-files, and bitmaps encoding intensity as greyscale. "
        "For the latter, the options 'rangex|y', and 'irange' have to be specified.",
        required=True,
        type=str,
    )
    arg_parser.add_argument(
        "--outputfile",
        "-o",
        default=None,
        dest="outfile",
        help="Destination file for generated csv document. Defaults to '<inputfile's basename>.csv.gz'.",
        required=False,
        type=str,
    )
    ## Options for bitmap conversion (postponed for v0.2 - TODO)
    # arg_parser.add_argument(
    #     "--xrange",
    #     "-x",
    #     default=None,
    #     dest="xrange",
    #     help="For bitmap conversion: range of input x-axis. Use X1>X2 if img-axis is reversed (left coord > right coord).",
    #     required=False,
    #     type=str,
    # )
    # arg_parser.add_argument(
    #     "--yrange",
    #     "-y",
    #     default=None,
    #     dest="yrange",
    #     help="For bitmap conversion: range of input y-axis. Use Y1>Y2 if img-axis is reversed (bottom coord > top coord)",
    #     required=False,
    #     type=str,
    # )
    # arg_parser.add_argument(
    #     "--irange",
    #     "-i",
    #     default=None,
    #     dest="irange",
    #     help="For bitmap conversion: range of input's intensities."
    #     " The greyscale values are mapped linearly onto the given range.",
    #     required=False,
    #     type=str,
    # )
    # arg_parser.add_argument(
    #     "--sqrt",
    #     action="store_true",
    #     default=False,
    #     dest="sqrt",
    #     help="Flag to indicate whether the given bitmap greyscale is sqrt-transformed.",
    #     required=False,
    # )
    arg_parser.add_argument(
        "--bindims",
        "-b",
        default="1,1",
        dest="bindims",
        help="Integer number of grid points on x- and y-axis to be summarized in NXxNY bins. "
        "Determines the resolution of the input as =~original_resolution./bindims. "
        "Note: high resolutions may entail long computation times.",
        required=False,
        type=str,
    )
    arg_parser.add_argument(
        "--override",
        "-w",
        action="store_true",
        default=False,
        dest="override",
        help="Flag to indicate whether any existing destination file should be overridden.",
        required=False,
    )
    arg_parser.add_argument(
        "--peaklist",
        action="store_true",
        default=False,
        dest="peaklist",
        help="Flag to indicate whether infile is a peak list. "
        "Such should be a table with headers indicating dimensions "
        "(e.g., '1H' and '13C'), which can be resolved by comparing to config.csv_column_name_links. "
        "If peak weights are included, add a column 'weight'. "
        "File must be readable by pandas.read_csv() with default arguments, see "
        "https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_csv.html",
        required=False,
    )
    return arg_parser.parse_args()


def get_input_type(file_ext):
    if file_ext == "ucsf":
        return InType.UCSF
    elif file_ext in ["txt", "txt.gz"]:
        return InType.BRUKER
    elif file_ext in ["csv", "csv.gz"]:
        return InType.PEAKLIST
    elif file_ext in ["png"]:
        return InType.BITMAP
    else:
        raise Exception(f"Couldn't derive file type from extension '{file_ext}'")


def process_args(args):
    processed_args = argparse.Namespace()
    wdir = Path(os.getcwd())
    True
    infile = Path(args.infile)
    if not infile.is_absolute():
        infile = wdir / infile
    if not infile.exists():
        return None, f"Input file '{infile}' does not exist."
    processed_args.infile = infile

    ext = infile.name.split(".")[-1]
    if ext == "gz":
        ext = ".".join(infile.name.split(".")[-2:])
    try:
        processed_args.type = get_input_type(ext)
    except:
        return None, f"Failed to map file extension '{ext}' to input type."

    if args.outfile:
        outfile = Path(args.outfile)
        if (outfile.name[-7:] != ".csv.gz") and (outfile.name[-4:] != ".csv"):
            return (
                None,
                f"Output file must have extension '.csv.gz' or '.csv'. Given: {outfile.name}",
            )
        if not outfile.is_absolute():
            outfile = wdir / outfile
    else:
        outfile = None

    if outfile is not None and outfile.exists() and not args.override:
        return (
            None,
            f"Output destination '{outfile}' already exists. You can force overriding by option --override.",
        )
    processed_args.outfile = outfile
    processed_args.override = args.override

    bindims = args.bindims
    if bindims is None:
        processed_args.bindims = None
    else:
        try:
            nX, nY = bindims.split(",")
            processed_args.bindims = int(nX), int(nY)
        except:
            return (
                None,
                f"Failed to process argument 'bindims', given value: {args.bindims}.",
            )

    ## Options for bitmap conversion (postponed for v0.2 - TODO)
    # bitmap_options = ["xrange", "yrange", "irange"]
    # if not processed_args.type == InType.BITMAP:
    #     for o in bitmap_options:
    #         if getattr(args, o) is not None:
    #             return (
    #                 None,
    #                 f"Option '{o}' is only accepted for the case of bitmap conversion.",
    #             )
    # else:
    #     for o in bitmap_options:
    #         val = getattr(args, o)
    #         try:
    #             x, y = val.split(",")
    #             setattr(processed_args, o, (float(x), float(y)))
    #         except:
    #             return None, f"Failed to process argument '{o}', given value: {val}."
    #
    # processed_args.sqrt = args.sqrt

    return processed_args, None


## Options for bitmap conversion (postponed for v0.2 - TODO)
# def bitmap2ptspec(args, name, smooth_radius=0.0, noise_cutoff=None):
#     with Image.open(args.infile) as img:
#         img.load()
#     # Grab first channel (shouldn't matter for grayscale)
#     # reverse 1st ax to have coord increase with index
#     data = np.flip(np.asarray(img)[:, :, 0], axis=0)
#
#     if np.max(data) > 1:
#         scale = 255
#     else:
#         scale = 1.0
#     data = args.irange[0] + (args.irange[1] - args.irange[0]) * data / scale
#     if args.sqrt:
#         data[data > 0] = data[data > 0] * data[data > 0]
#         data[data < 0] = -data[data < 0] * data[data < 0]
#
#     if args.yrange[0] > args.yrange[1]:
#         # Image has reversed y-axis
#         data = np.flip(data, axis=0)
#     if args.xrange[0] > args.xrange[1]:
#         # Image has reversed x-axis
#         data = np.flip(data, axis=1)
#     xrange = min(args.xrange), max(args.xrange)
#     yrange = min(args.yrange), max(args.yrange)
#     header = dict(FRanges=(xrange, yrange))
#
#     grid_spec = RasterSpectrum(
#         data,
#         header,
#         name=name,
#         denoising_pars=dict(smoothRadius=smooth_radius, noiseFilterLevel=noise_cutoff),
#     )
#
#     plotRasterSpectrum(
#         grid_spec,
#         title="Test " + name,
#         clim=(-np.max(data), np.max(data)),
#         cmap="gray",
#         show=True,
#     )
#
#     pt_spec = make_point_spectrum(grid_spec, args)
#     return pt_spec


def ucsf2ptspec(args, name, smooth_radius=0.0, noise_cutoff=None):
    matrix, header = parseUCSF(args.infile, args.infile.parent)
    grid_spec = RasterSpectrum(
        matrix,
        header,
        name=name,
        denoising_pars=dict(smoothRadius=smooth_radius, noiseFilterLevel=noise_cutoff),
    )
    pt_spec = make_point_spectrum(grid_spec, args)
    return pt_spec


def bruker2ptspec(args, name, smooth_radius=0.0, noise_cutoff=None):
    fullData, header = parseBrukerSpectrum(args.infile)
    grid_spec = RasterSpectrum(
        fullData,
        header,
        name=name,
        denoising_pars=dict(smoothRadius=smooth_radius, noiseFilterLevel=noise_cutoff),
    )
    pt_spec = make_point_spectrum(grid_spec, args)
    return pt_spec


def peaklist2ptspec(args, name):
    print(f"Parsing peak list from '{args.infile}'")
    df = pd.read_csv(args.infile)
    df = df.rename(columns={c: c.strip().lower() for c in df.columns})

    candis = {k: [e.lower() for e in v] for k, v in csv_column_name_links.items()}
    # For character inclusion test:
    # col_char[c] in df.column[i] => col[c] = i
    col_char = {
        "1h": "h",
        "13c": "c",
    }

    cols = {c: None for c in candis}
    for c in candis:
        for col in df.columns:
            if col in candis[c]:
                cols[c] = col
                break
        if cols[c] is None and c in col_char:
            for col in df.columns:
                # Check if this column was already used for any other dimension
                unique = not np.any(
                    [
                        col_char[c] in col
                        for col in df.columns
                        if col not in cols.values()
                    ]
                )
                if unique and col_char[c] in col:
                    cols[c] = col
                    break
        if cols[c] is None:
            print(f"Found no column for {c}.")
        else:
            print(f"Using column {cols[c]} for {c}.")

    data = {c: np.array(df.loc[:, cols[c]]) for c in cols if cols[c] is not None}
    if cols["weight"] is None:
        data["weight"] = np.ones(len(df))

    if "1H" not in data:
        print(f"Couldn't find 1H-column in peak list '{args.infile}'")
        sys.exit(1)
    if "13C" not in data:
        print(f"Couldn't find 13C-column in peak list '{args.infile}'")
        sys.exit(1)
    coords = np.vstack((np.astype(data["13C"], float), np.astype(data["1H"], float))).T
    ptspec = PointSpectrum(coords, weights=data["weight"], name=name)
    return ptspec


def make_point_spectrum(grid_spec, args):
    # Determine binning dimensions
    if args.bindims is None:
        # use input resolution
        nbin = (1, 1)
    else:
        nbin = args.bindims
        nbin = np.maximum(nbin, 1)
        name = grid_spec.name + f"{nbin}"
    coords, weights, pointClusterArea = grid_spec.getRasterDataAsPeaks(
        nbin=nbin, xVar="F1", signalThreshold=None, cutGridToBin=True
    )
    pt_spec = PointSpectrum(
        coords,
        weights,
        spec=grid_spec,
        fromRaster=True,
        binning=nbin,
        pointArea=pointClusterArea,
        name=name,
    )
    return pt_spec


def write_ptspec_as_csv(pt_spec, fn):
    if fn.name[-3:] == ".gz":
        f = gzip.open(fn, "wb")
    else:
        f = open(fn, "w")
    df = {
        "weight": pt_spec.weights,
        "1H": pt_spec.coords[:, 1],
        "13C": pt_spec.coords[:, 0],
    }
    pd.DataFrame(df).to_csv(f, sep=",")
    f.close()
    print(f"\nWritten data to {fn}.\n")


def run(args):
    infile_split = args.infile.name.split(".")
    if infile_split[-1] == "gz":
        name = ".".join(infile_split[:-2])
    else:
        name = ".".join(infile_split[:-1])
    if args.type == InType.BITMAP:
        ## Options for bitmap conversion (postponed for v0.2 - TODO)
        # pt_spec = bitmap2ptspec(args, name)
        raise Exception("Sorry, bitmap-to-spectrum conversion not implemented, yet.")
    elif args.type == InType.BRUKER:
        pt_spec = bruker2ptspec(args, name)
    elif args.type == InType.UCSF:
        pt_spec = ucsf2ptspec(args, name)
    elif args.type == InType.PEAKLIST:
        pt_spec = peaklist2ptspec(args, name)
    else:
        assert False

    # Determine outfile name
    if args.outfile is None:
        wdir = Path(os.getcwd())
        args.outfile = wdir / (name + ".csv.gz")
    if args.outfile.exists() and not args.override:
        print(f"Failed to write output to '{args.outfile}'.")
        print("File already exists. Force overriding with option '--override'.")
        sys.exit(1)

    # Save csv
    write_ptspec_as_csv(pt_spec, args.outfile)


def main():
    args = parse_args()
    processed_args, msg = process_args(args)
    if processed_args is None:
        print("\nspec2csv failed with message:\n" + msg)
        sys.exit(1)
    run(processed_args)


if __name__ == "__main__":
    main()



import numpy as np
from pprint import pformat
import scipy

from mcfnmr.utils.geometry import RectRegion
from itertools import product


class PointSpectrum(object):
    def __init__(
        self,
        coords,
        weights,
        spec=None,
        fromRaster=False,
        binning=None,
        pointArea=None,
        name=None,
        regions=[],
    ):
        assert not (spec is None and fromRaster)
        assert not (binning is None and fromRaster)
        assert not (pointArea is None and fromRaster)
        if len(weights) != len(coords):
            raise Exception(f"Length of weights != length of coords for '{name}'")
        if np.any(np.isnan(weights)):
            raise Exception(f"NaN in weight array for '{name}'")
        self.coords = np.array(coords)
        self.coords.shape = (self.coords.shape[0], 2)
        if len(weights) != len(coords):
            raise Exception(
                f"Length of weights != length of coords for '{name}' after reshape..."
            )
        self.weights = np.array(weights).flatten()
        self.unknownRegions = []
        self.representsRasterData = fromRaster
        self.binning = binning
        self.pointDimensions = None
        self.pointArea = pointArea
        self.name = name
        self.regions = regions

        if spec is not None:
            self.scanNr = spec.scanNr
            self.noiseStd = spec.noiseStd
            self.regions = spec.regions
            self.pointDimensions = (
                int(spec.pointDimensions[0] / binning[0]),
                int(spec.pointDimensions[1] / binning[1]),
            )
            self.FRanges = spec.FRanges

        if self.coords.size == 0:
            self.coords.shape = (0, 2)
        assert self.coords.size == 2 * self.weights.size

    def isPointSpectrum(self):
        return True

    def size(self):
        return len(self.weights)

    def __str__(self):
        s = "Pointspectrum (size: %d)\n" % self.size()
        s += "  weights:\n"
        s += pformat(self.weights, indent=3)
        s += "\n  coords:\n"
        s += pformat(self.coords, indent=3)
        return s

    def mult(self, x):
        return PointSpectrum(self.coords, x * self.weights)

    def add(self, spec):
        return PointSpectrum(
            np.concatenate((self.coords, spec.coords)),
            np.concatenate((self.weights, spec.weights)),
        )

    def isEmpty(self):
        return self.weights.size == 0


def emptyPointSpectrum():
    return PointSpectrum([], [])


def makeMix(basis, alpha=None, name=None):
    if alpha is None:
        alpha = {c: 1.0 for c in basis}
    # Create mixture
    X = emptyPointSpectrum()
    for k in alpha.keys():
        if k not in basis:
            print("Key '%s' not in basis. Ignoring..." % k)
            continue
        b = basis[k]
        if np.any(np.isnan(b.weights)):
            raise Exception("Malformed intensity for compound %s!" % k)
        X = X.add(b.mult(alpha[k]))
    if name is not None:
        X.name = name
    return X


def makeRasterPointSpectrum(rasterspec, nbin=(1, 1), signalThreshold=None, xVar="F1"):
    coords, weights, pointClusterArea = rasterspec.getRasterDataAsPeaks(
        nbin=nbin, signalThreshold=signalThreshold, xVar=xVar
    )
    spec = PointSpectrum(
        coords,
        weights,
        spec=rasterspec,
        fromRaster=True,
        binning=nbin,
        pointArea=pointClusterArea,
        name=rasterspec.name,
    )
    return spec


def constructRasterFromPointSpectrum(spec, rasterDim, rasterRange, kernelWidths):
    range0 = rasterRange[0][1] - rasterRange[0][0]
    range1 = rasterRange[1][1] - rasterRange[1][0]
    dy, dx = range0 / rasterDim[0], range1 / rasterDim[1]

    # Construct kernel to be added to spectrum around coords in Z
    # We assume that the kernel width corresponds to 4 stds of
    # the corresponding normal distribution. For the kernel data, we
    # keep 6 stds
    kernelDims = [
        int(np.ceil(kernelWidths[1] * 3 / dy)),
        int(np.ceil(kernelWidths[0] * 3 / dx)),
    ]
    # Assume have odd dimension for centering
    if kernelDims[0] % 2 == 0:
        kernelDims[0] += 1
    if kernelDims[1] % 2 == 0:
        kernelDims[1] += 1
    # Central point index in kernel
    ix0 = ((kernelDims[0] - 1) / 2, (kernelDims[1] - 1) / 2)

    kernelRelY = (np.arange(kernelDims[0]) - ix0[0]) * dx
    kernelRelX = (np.arange(kernelDims[1]) - ix0[1]) * dy
    kernelRelX, kernelRelY = np.meshgrid(kernelRelX, kernelRelY)
    vary = kernelWidths[0] * kernelWidths[0] / 16
    varx = kernelWidths[1] * kernelWidths[1] / 16
    d = scipy.stats.multivariate_normal(mean=[0.0, 0.0], cov=[[vary, 0], [0, varx]])
    kernelRelXY = np.array(list(zip(kernelRelY.flatten(), kernelRelX.flatten())))
    kernel = d.pdf(kernelRelXY)
    kernel.shape = kernelRelX.shape
    # Make shape round
    ks = kernel.shape
    for i0 in range(ks[0]):
        for i1 in range(ks[1]):
            if (
                np.linalg.norm((2 * (i0 - ix0[0]) / ks[0], 2 * (i1 - ix0[1]) / ks[1]))
                > 1
            ):
                kernel[i0, i1] = 0

    # Debug
    # pcm = plt.pcolormesh(kernelRelY, kernelRelX, kernel, cmap="hot")
    # plt.show()

    # make grid of pixel centerpoints
    yspan = np.linspace(
        rasterRange[0][0] + dy / 2, rasterRange[0][1] - dy / 2, rasterDim[0]
    )
    xspan = np.linspace(
        rasterRange[1][0] + dx / 2, rasterRange[1][1] - dx / 2, rasterDim[1]
    )
    X, Y = np.meshgrid(xspan, yspan)
    Z = np.zeros_like(X)

    # Add kernels concentrated at single peaks
    ixrad = int((kernelDims[0] - 1) / 2), int((kernelDims[1] - 1) / 2)
    region = RectRegion(xrange=rasterRange[0], yrange=rasterRange[1])
    for xy, w in zip(spec.coords, spec.weights):
        if not region.hitTest(xy):
            # Ignore peaks outside of target range
            continue
        # Find closest coord index for xy
        yix = int(np.round((rasterDim[0] - 1) * (xy[0] - rasterRange[0][0]) / range0))
        xix = int(np.round((rasterDim[1] - 1) * (xy[1] - rasterRange[1][0]) / range1))
        yix0, yix1 = max(yix - ixrad[0], 0), min(yix + ixrad[0] + 1, rasterDim[0])
        xix0, xix1 = max(xix - ixrad[1], 0), min(xix + ixrad[1] + 1, rasterDim[1])
        kyix0, kyix1 = max(ixrad[0] - yix, 0), kernelDims[0] - max(
            (yix + ixrad[0]) - (rasterDim[0] - 1), 0
        )
        kxix0, kxix1 = max(ixrad[1] - xix, 0), kernelDims[1] - max(
            (xix + ixrad[1]) - (rasterDim[1] - 1), 0
        )
        Z[yix0:yix1, xix0:xix1] += w * kernel[kyix0:kyix1, kxix0:kxix1]
    return X, Y, Z


class Indexer2D:
    def __init__(self, rowfirst, nX, nY, xdir, ydir, rangex, rangey):
        self.rowfirst=rowfirst 
        self.nX=nX
        self.nY=nY
        self.xdir=xdir 
        self.ydir=ydir
        # Determine coefficients for index function
        self.x_ix0 = 0 if xdir == 1 else nX - 1
        self.y_ix0 = 0 if ydir == 1 else nY - 1
        self.x_ixstep = 1 if rowfirst else nY
        self.y_ixstep = nX if rowfirst else 1
        # Determine coefficients for position2index
        self.dx = xdir*(rangex[1]-rangex[0])/nX
        self.dy = ydir*(rangey[1]-rangey[0])/nY
        self.offsetx = rangex[0] if xdir==1 else rangex[1]
        self.offsety = rangey[0] if ydir==1 else rangey[1]
        
    def __call__(self, i, j):
        return (self.y_ixstep * (self.y_ix0 + self.ydir * j) 
                + self.x_ixstep * (self.x_ix0 + self.xdir * i))

    def getIndex(self, v, dv, offset):
        return int(np.round((v - offset)/dv))

    def getXIndex(self, x):
        return self.getIndex(x, self.dx, self.offsetx)
    
    def getYIndex(self, y):
        return self.getIndex(y, self.dy, self.offsety)
        
    def getNeighs(self, xy, rx, ry):
        # Return indices of neighbors of point xy within ranges rx, ry
        x, y = xy
        # Index base
        ixx, ixy = self.getXIndex(x), self.getYIndex(y)
        # randes converted to inices
        nrx, nry = int(np.ceil(np.abs(rx/self.dx)))+1, int(np.ceil(np.abs(ry/self.dy)))+1
        # neigh index ranges
        r_ixx, r_ixy = np.arange(ixx-nrx, ixx+nrx+1), np.arange(ixy-nry, ixy+nry+1)
        r_ixx = r_ixx[(0 <= r_ixx) & (r_ixx < self.nX)]
        r_ixy = r_ixy[(0 <= r_ixy) & (r_ixy < self.nY)]
        ixs = np.array([self(i,j) for i,j in product(r_ixx, r_ixy)])
        return ixs
    
        
    def makeMatrix(self, weights):    
        # Copy weight values into matrix
        Z = np.zeros((self.nX, self.nY))
        for i in range(self.nX):
            for j in range(self.nY):
                ix_ij = self(i,j)
                Z[i, j] = weights[ix_ij]
        return Z
        
    
def detect_grid(coords):
    # Try to derive the grid dimensions and ranges
    # from the given coordinate array
    x, y = coords[:,0], coords[:,1]
    rangex, rangey = (min(x), max(x)), (min(y), max(y))
    if rangex[0] == rangex[1] or rangey[0] == rangey[1]:
        # No variation of coords in one direction
        return None, "No variation of coords in one direction"
    dx, dy = x[1:] - x[:-1], y[1:] - y[:-1]
    if dx[0] != 0 and dy[0] != 0:
        # The first step of a serialized grid should only
        # increase one coordinate
        return (
            None,
            "The first step of a serialized grid should only increase one coordinate ",
        )

    # Determine index order of serialization
    if dx[0] != 0 and dy[0] == 0:
        rowfirst = True
        unique_dy = set(np.round(h, 12) for h in set(dy) if h != 0.0)
        unique_dx = set(np.round(h, 12) for h in set(dx) if h != 0.0)
        # dx sometimes jumps back from max to zero
        maxdx = np.max(np.abs(unique_dx))
        unique_dx.difference_update((maxdx, -maxdx))
    else:
        rowfirst = False
        unique_dx = set(np.round(h, 12) for h in set(dx) if h != 0.0)
        unique_dy = set(np.round(h, 12) for h in set(dy) if h != 0.0)
        # dy sometimes jumps back from max to zero
        maxdy = np.max(np.abs(list(unique_dy)))
        unique_dy.difference_update((maxdy, -maxdy))

    # Find number of grid points on axes
    if len(unique_dx) > 1:
        # Variable distances of steps in x-direction
        return None, "Variable distances of steps in x-direction"
    if len(unique_dy) > 1:
        # Variable distances of steps in y-direction
        return None, "Variable distances of steps in y-direction"

    dx0, dy0 = unique_dx.pop(), unique_dy.pop()

    # Determine number of grid points on axes
    X, Y = sorted(set(x)), sorted(set(y))
    if len(X) * len(Y) != len(coords):
        return (
            None,
            "Total number of points is different from nr of x-coords times nr of y-coords",
        )
    
    nX, nY = len(X), len(Y)
    xdir, ydir = int(np.sign(dx0)), int(np.sign(dy0))
    indexer = Indexer2D(rowfirst, nX, nY, xdir, ydir, rangex, rangey)
    return X, Y, indexer



import os
import gzip
import numpy as np
import struct
import subprocess as sp

from mcfnmr.config import DEBUG, UCSFDATA_BIN


def parseUCSF(fn, outdir, cleanup=True):
    # Generate and process header

    headerFile = outdir / (fn.name + "_header.txt")
    header = getUCSFHeader(headerFile, fn)

    # Generate and process matrix
    binaryFile = outdir / (fn.name + "_matrix.m")
    matrix = getUCSFMatrix(binaryFile, fn, header)

    if cleanup:
        os.remove(headerFile)
        os.remove(binaryFile)

    return matrix, header


def parseUCSFHeader(fn):
    with open(fn, "r") as f:
        lines = f.readlines()
    header = {}
    for l in lines[2:]:
        ll = l.strip().split()
        k = ll[0].strip()
        for s in ll[1:-2]:
            k += " " + s.strip()
        header[k] = (float(ll[-2]), float(ll[-1]))
    if DEBUG > 1:
        print(header)
    # Add frequency ranges to header
    C_range = (header["upfield ppm"][0], header["downfield ppm"][0])
    H_range = (header["upfield ppm"][1], header["downfield ppm"][1])
    header["FRanges"] = (C_range, H_range)
    return header


def getUCSFHeader(headerFile, ucsfFile):
    with open(headerFile, "w") as f:
        sp.call([UCSFDATA_BIN, ucsfFile], stdout=f)
    header = parseUCSFHeader(headerFile)
    return header


def getUCSFMatrix(dataFile, ucsfFile, header):
    # "If the -m flag is present the data matrix is written in binary format as IEEE floats with your machine's native byte order.
    #  The matrix is written out with highest axis varying fastest."
    with open(dataFile, "w") as f:
        sp.call([UCSFDATA_BIN, "-m", ucsfFile], stdout=f)

    file = open(dataFile, "rb")
    nbytes = os.path.getsize(dataFile)
    n = int(nbytes / 4)
    assert n == nbytes / 4
    f = struct.unpack("f" * n, file.read(4 * n))
    f = list(reversed(f))

    mDim = header["matrix size"]
    mDim = (int(mDim[0]), int(mDim[1]))
    if DEBUG > 1:
        print("len(f)=", len(f))
        print("matrix size=", mDim[0] * mDim[1])

    assert mDim[0] * mDim[1] == len(f)
    f = np.array(f, dtype=np.float32)
    f.shape = mDim

    return f


def parseBrukerSpectrum(fn):
    # Specific lines that indicate the end of the header for the used txt formats
    headerStopSymbols = ["= Re Matrix", "= Processed Matrix", "= RR Matrix"]

    def headerEnd(line):
        for s in headerStopSymbols:
            if line[: len(s)] == s:
                return True
        return False

    headerLines = []
    with gzip.open(fn, "rb") as f:
        # Load header
        headerOpen = True
        while headerOpen:
            l = f.readline()[:-1].decode()
            headerLines.append(l)
            if headerEnd(l):
                headerOpen = False
                continue
            print(l[:-1])
        header = parseBrukerHeader(headerLines)
        print("Loaded header data for '%s'" % fn.name)

        # Read matrix data
        pointDimensions = header["pointDimensions"]
        dataStr = []
        for i in range(pointDimensions[0]):
            # print("%d/%d"%(i,self.pointDimensions[0]))
            l = f.readline().decode()
            dataStr.append(l)
    fullData = np.loadtxt(dataStr, delimiter=" ")
    print("Loaded spectrum data for '%s'" % fn.name)
    assert fullData.shape == pointDimensions

    return fullData, header


def parseBrukerHeader(lines):
    # float variable are 2D-header pairs, holding numeric information
    floatVariables = {"SweepWidth", "Frequency", "Offset", "PointsCount"}
    intVariables = {"ExpCount"}
    headerDict = dict()
    for l in lines:
        if l.count("=") == 1:
            # key value pair, probably
            k, v = l.split("=")
            k, v = k.strip(), v.strip()
            if k in floatVariables:
                v = v.split(" ")
                v = float(v[0]), float(v[1])
            if k in intVariables:
                v = int(v)
            headerDict[k] = v

    # Calculate spectral dimansions from header information and include into header
    headerDict["scanNr"] = headerDict.get("ExpCount", None)
    headerDict["pointDimensions"] = tuple(
        [int(n) for n in reversed(headerDict["PointsCount"])]
    )
    range1 = np.array(
        (
            headerDict["Offset"][0] - 0.5 * headerDict["SweepWidth"][0],
            headerDict["Offset"][0] + 0.5 * headerDict["SweepWidth"][0],
        )
    )
    range2 = np.array(
        (
            headerDict["Offset"][1] - 0.5 * headerDict["SweepWidth"][1],
            headerDict["Offset"][1] + 0.5 * headerDict["SweepWidth"][1],
        )
    )
    range1 /= headerDict["Frequency"][0]
    range2 /= headerDict["Frequency"][1]
    headerDict["FRanges"] = (range2, range1)

    return headerDict



from pprint import pprint
from collections import defaultdict
import numpy as np
import pickle
import os

from mcfnmr.routines.utils import (
    prepareParameters,
    makeResultsDF,
    cache_dir,
    figs_dir_oldb,
)
from mcfnmr.demodata.loading import loadOLDBCompoundLib
from mcfnmr.routines.quantification import (
    runSingleQuantificationIncremental,
    runSingleQuantification,
    makeSetupStr,
    showCompounds,
)
from mcfnmr.demodata import OLDBCompoundWeights
from mcfnmr.routines.plotting import (
    plotAssignmentStatsOLDB,
    plotErrorStats,
    plotPredictionDiagonals,
)
from mcfnmr.config import OUTDIR


def runRange(pars, arSpan, lib=None, load=True):
    incrementalFit = pars["incrementalFit"]
    if incrementalFit:
        results = runSingleQuantificationIncremental(
            pars=pars, arSpan=arSpan, lib=lib, load=load, return_all=True
        )
    else:
        results = []
        for ar in arSpan:
            res = runSingleQuantification(pars, ar=ar, lib=lib, load=load)
            results.append(res)
    return results


def main(cfg, show=False, loadtmp=False):
    
    tmpfn = OUTDIR / "tmp" / "comparison_quantification_all_dfCompounds.pickle"
    if not (OUTDIR / "tmp").exists():
        os.makedirs(OUTDIR / "tmp")
        print(f"Created directory '{OUTDIR / 'tmp'}'")
    
    assert cfg["task"] == "comparison_quantification"
    assert cfg["lib"].lower() == "inhouse"
    mix = cfg["mix"].split(".")
    assert len(mix) == 2
    series, experiment = mix
    assert series in ["I", "II", "III", "all"]
    assert experiment in ["a", "b", "c", "all"]
    assert cfg["setup"] == "all"  # This iterates through setups A, B, and C, below

    # For incremental assignment
    arSpan = np.array(
        [
            0.025,
            0.05,
            0.075,
            0.1,
            0.125,
            0.15,
            0.175,
            0.2,
            0.225,
            0.25,
            0.3,
            0.35,
            0.4,
            0.5,
        ]
    )

    plot = cfg.get("plot", False)


    if loadtmp and tmpfn.exists():
        with open(tmpfn, "rb") as f:
            all_dfCompounds, data = pickle.load(f)
        print(f"Loaded all_dfCompounds from {tmpfn}")
    else:
        # OLDB compound lib
        lib = loadOLDBCompoundLib()
    
        # All individual compounds with concentrations 30mM
        libID = "OLDBcompoundLib"
        targetSpecID = series + experiment + "_01"
        binning = (2, 4)
        noiseFilterLevel = None
        smoothing = 0
        pars = prepareParameters(
            isolated_fit="all",
            targetSpecID=targetSpecID,
            pointTarget=False,
            libID=libID,
            binning=binning,
            specDomain="inhouse",
            incrementalFit="all",
            noiseFilterLevel=None,
            smoothing=0,
            workingDir=cfg["outdir"],
            recompute=cfg.get("recompute", True),
        )
        mixes = [
            "Ia_01",
            "Ib_01",
            "Ic_01",
            "IIa_01",
            "IIb_01",
            "IIc_01",
            "IIIa_01",
            "IIIb_01",
            "IIIc_01",
        ]
        # Collecting all into one dataframe
        data = defaultdict(list)
        all_dfCompounds = defaultdict(dict)
        for isoFit, incrFit in ((False, False), (False, True), (True, False)):
            for mixID in mixes:
                targetSpecID = mixID
                pars.update(
                    dict(
                        isolated_fit=isoFit,
                        targetSpecID=targetSpecID,
                        incrementalFit=incrFit,
                    )
                )
                results = runRange(pars, arSpan, lib, load=not cfg["recompute"])
    
                # This assumes that all weight of X is in the target regions
                dfTargets, dfCompounds = makeResultsDF(
                    results, intensityCorrection=True, originalWeights=OLDBCompoundWeights
                )
                all_dfCompounds[(isoFit, incrFit)][mixID] = dfCompounds
    
                restrict_arSpan = cfg.get("ar", None)
                if restrict_arSpan:
                    dfTargets = dfTargets.iloc[
                        np.isin(dfTargets["assignment radius"], restrict_arSpan)
                    ]
                    dfCompounds = dfCompounds.iloc[
                        np.isin(dfCompounds["assignment radius"], restrict_arSpan)
                    ]
                else:
                    restrict_arSpan = arSpan
    
                setupStr = makeSetupStr(
                    targetSpecID,
                    binning,
                    None,
                    None,
                    None,
                    libID,
                    restrict_arSpan,
                    noiseFilterLevel,
                    smoothing,
                )
                targetSpecID = pars["targetSpecID"]
                stats = plotAssignmentStatsOLDB(
                    dfCompounds,
                    dfTargets,
                    libID,
                    targetSpecID,
                    showCompounds,
                    outdir=figs_dir_oldb,
                    show=False,
                    setupStr=setupStr,
                )
                print(f'error_stats["{targetSpecID}"] = \\')
                pprint(stats["errors"])
                errs = stats["errors"]
                for k, vals in errs.items():
                    data[k].extend(vals)
        
        with open(tmpfn, "wb") as f:
            pickle.dump((all_dfCompounds, data), f)
        print(f"Saved all_dfCompounds to '{tmpfn}'")
        
    if plot:
        # All setups, hue by setup
        plot_ar = 0.15
        plotPredictionDiagonals(all_dfCompounds, ar=plot_ar, show=show)
        for isoFit, incrFit in ((False, False), (False, True), (True, False)):
            # Single setups, hue by compound specific signal to noise
            plotPredictionDiagonals(
                all_dfCompounds, ar=plot_ar, setup=(isoFit, incrFit), show=show
            )

        if cfg.get("ar", None) is None:
            cfg["ar"] = arSpan
        ars = [float(ar) for ar in cfg["ar"]]
        figname = "quantitativeReconstructionErrors" + str(ars)
        plotErrorStats(data, figname, show=show)


if __name__ == "__main__":
    # Test run
    cfg = dict(
        lib="InHouse",
        mix="all.all",
        recompute=False,
        task="comparison_quantification",
        setup="all",
        # Assignment radii to compute (ar=None shows all)
        ar=None,
        plot=True,
        outdir=cache_dir,
    )

    main(cfg, show=True, loadtmp=False)



from copy import deepcopy


# Different libraries provided by MM
metaboMinerLibNames = [
    "Biofluid ( all )",
    "Biofluid ( common )",
    "CSF ( all )",
    "CSF ( common )",
    "Plasma ( all )",
    "Plasma ( common )",
    "Urine ( all )",
    "Urine ( common )",
]
metaboMinerLibNameDict = {
    "all" : "Biofluid ( all )",
    "common": "Biofluid ( common )",
    "csf-all": "CSF ( all )",
    "csf-common": "CSF ( common )",
    "plasma-all": "Plasma ( all )",
    "plasma-common": "Plasma ( common )",
    "urine-all": "Urine ( all )",
    "urine-common": "Urine ( common )",
}

# Subsets of MaetaboMiner mixes
mix_selections = {
    "ph7": ["N925", "N926", "N987", "N988"],
    "SMART-Miner": ["N925", "N987", "N988"],
    "allMM": ["N880", "N907", "N925", "N926", "N987", "N988"],
}

# Bin sizes for raster data
binning = {
    "N907": (1, 4),
    "N925": (2, 4),
    "N987": (1, 2),
    "N926": (1, 2),
    "N880": (1, 2),
    "N988": (1, 2),
}


# Template & defaults
caseSpec0 = dict(
    name="default",
    # not using peak weights for comparison as this is not used by metabominer, either
    uniformPeaks=True,
    # Whether to use the point spectrum or the raster spectrum as target
    pointTarget=True,
    # Whether to fit compounds individually
    isolated_fit=False,
    # Should this fit build on previous ones (with lower assingment radius)?
    incrementalFit=False,
    # MetaboMiner Compound lib
    libID="metabominer",
    mixIDs=["N925", "N987", "N988"],  # SMART-Miner mixes
)

caseSpec1 = deepcopy(caseSpec0)
caseSpec1["name"] = "base case"

## Setups for Paper
# A: joint, single pass, grid
caseSpecA = deepcopy(caseSpec1)
caseSpecA["pointTarget"] = False
caseSpecA["name"] = "setup A"

# B: independent, single pass, grid
caseSpecB = deepcopy(caseSpecA)
caseSpecB["isolated_fit"] = True
caseSpecB["name"] = "setup B"

# C: joint, incremental, grid
caseSpecC = deepcopy(caseSpecA)
caseSpecC["incrementalFit"] = True
caseSpecC["name"] = "setup C"

# D: independent, single pass, peak list
caseSpecD = deepcopy(caseSpecA)
caseSpecD["pointTarget"] = True
caseSpecD["name"] = "setup D"

# E: joint, incremental, peak list
caseSpecE = deepcopy(caseSpecA)
caseSpecE["pointTarget"] = True
caseSpecE["incrementalFit"] = True
caseSpecE["name"] = "setup E"

# E: joint, incremental, peak list
caseSpecF = deepcopy(caseSpecA)
caseSpecF["pointTarget"] = True
caseSpecF["isolated_fit"] = True
caseSpecF["name"] = "setup F"

baseSpec = dict(
    A=caseSpecA, B=caseSpecB, C=caseSpecC, D=caseSpecD, E=caseSpecE, F=caseSpecF
)


# For individual mixes and setups from paper
# map setup -> mix -> pars
case_dict = dict(A={}, B={}, C={}, D={}, E={}, F={})
for setup in ["A", "B", "C", "D", "E", "F"]:
    for mix in ["N907", "N925", "N926", "N987", "N988", "N880", "all"]:
        mix_list = ["N925", "N987", "N988"] if mix == "all" else [mix]
        case_dict[setup][mix] = deepcopy(baseSpec[setup])
        case_dict[setup][mix]["mixIDs"] = mix_list

## Make case dicts for A, B, C for quantification single runs (no multiple mixes allowed)
for series in ["I", "II", "III"]:
    for experiment in ["a", "b", "c"]:
        for setup in ["A", "B", "C"]:
            mixID = series + "." + experiment
            case_dict[setup][mixID] = deepcopy(baseSpec[setup])
            case_dict[setup][mixID]["mixIDs"] = [mixID]

# Map compound full names of  OLDB-lib to abbreviations
compoundShortNames = {
    "Raffinose": "Raf",
    "Pimelic acid": "Pim",
    "D-Xylose": "Xyl",
    "L-Cysteine": "Cys",
    "Nicotinic acid": "Nic",
    "Malonic acid": "Mal",
    "L-Proline": "Pro",
    "Anthranilic acid": "Ant",
    "Citric acid": "Cit",
    "L-Rhamnose": "Rha",
    "1,2,4-Benzenetricarboxylic acid": "Bnz4",
    "1,2,3-Benzenetricarboxylic acid": "Bnz3",
    "D-Mannose": "Man",
    "L-Glutamine": "Glu",
    "4-Phenylbutanoic acid": "PheB",
    "Succinic acid": "Suc",
    "Benzoic acid": "Bnz",
    "Octanoic acid": "Oct",
    "Biotin": "Bio",
    "4-Hydroxybenzoic acid": "HBz",
    "9-Anthracenecarboxylic acid": "AntCx",
    "L-Isoleucine": "ILeu",
    "L-Tartaric acid": "Tar",
    "L-Tyrosine": "Tyr",
    "L-Tryptophane": "Try",
    "L-Glutamic acid": "GluA",
    "L-Methionine": "Met",
    "L-Phenylalanine": "PheA",
    "4-Hydroxycinnamic acid": "HCin",
    "D-Glucose": "Gluc",
    "L-Leucine": "LLeu",
    "(E)-Cinnamic acid": "CinA",
    "Tetradecanoic acid": "TDec",
    "Vanillic acid": "Van",
}

# Inverse map
compoundLongNames = dict([(v, k) for k, v in compoundShortNames.items()])

# Updated short names
textbook_abbreviations = {
    "Raffinose": "Raf",
    "Pimelic acid": "PA",
    "D-Xylose": "Xyl",
    "L-Cysteine": "Cys",
    "Nicotinic acid": "NA",
    "Malonic acid": "MA",
    "L-Proline": "Pro",
    "Anthranilic acid": "AA",
    "Citric acid": "CA",
    "L-Rhamnose": "Rha",
    "1,2,4-Benzenetricarboxylic acid": "1,2,4-BTCA",
    "1,2,3-Benzenetricarboxylic acid": "1,2,3-BTCA",
    "D-Mannose": "Man",
    "L-Glutamine": "Gln",
    "4-Phenylbutanoic acid": "4-PBA",
    "Succinic acid": "SA",
    "Benzoic acid": "BA",
    "Octanoic acid": "OA",
    "Biotin": "Biotin",
    "4-Hydroxybenzoic acid": "4-HBA",
    "9-Anthracenecarboxylic acid": "9-ACA",
    "L-Isoleucine": "Ile",
    "L-Tartaric acid": "TA",
    "L-Tyrosine": "Tyr",
    "L-Tryptophane": "Trp",
    "L-Glutamic acid": "Glu",
    "L-Methionine": "Met",
    "L-Phenylalanine": "Phe",
    "4-Hydroxycinnamic acid": "4-HCA",
    "D-Glucose": "Glc",
    "L-Leucine": "Leu",
    "(E)-Cinnamic acid": "tCA",
    "Tetradecanoic acid": "TDA",
    "Vanillic acid": "VA",
}

# Some substances have aliases
correctSubstrateNames = {
    "1,2,3-Benzentricarboxylic acid": "1,2,3-Benzenetricarboxylic acid",
    "1,2,4-Benzentricarboxylic acid": "1,2,4-Benzenetricarboxylic acid",
}


def correctShortCompoundLabels(compound_list):
    res = []
    for c in compound_list:
        clong = compoundLongNames.get(c, c)
        cnew = textbook_abbreviations[clong]
        res.append(cnew)
    return res


# Map mix identifiers to descriptive names
mixNamesOLDB = {}
for s1 in ["I", "II", "III"]:
    for s2 in ["a", "b", "c"]:
        mixNamesOLDB[s1 + s2 + "_01"] = s1 + s2 + "_01"
        mixNamesOLDB[s1 + s2] = s1 + s2

# Number of NMR scans by which the OLDB spectra were recorded
# Intensity scales linearly with the scan number (used for
# renormalization to predict the concentration)
scanNumberIndividualCompounds = 4
scanNumbersMixes = {
    "Ia_01": 8,
    "Ia_02": 8,
    "Ia_03": 16,
    "Ia_04": 16,
    "Ib_01": 16,
    "Ic_01": 64,
    "IIa_01": 4,
    "IIa_03": 4,
    "IIa_04": 4,
    "IIb_01": 8,
    "IIc_01": 16,
    "IIIa_01": 8,
    "IIIb_01": 64,
    "IIIc_01": 32,
}


intensityScalings = {
    "(E)-Cinnamic acid": 3,
    "1,2,3-Benzenetricarboxylic acid": 2,
    "1,2,4-Benzenetricarboxylic acid": 1,
    "4-Hydroxybenzoic acid": 3,
    "4-Hydroxycinnamic acid": 3,
    "4-Phenylbutanoic acid": 3,
    "9-Anthracenecarboxylic acid": 3,
    "Anthranilic acid": 2,
    "Benzoic acid": 2,
    "Biotin": 3,
    "Citric acid": 2,
    "D-Glucose": 2,
    "D-Mannose": 3,
    "D-Xylose": 2,
    "L-Cysteine": 2,
    "L-Glutamic acid": 2,
    "L-Glutamine": 2,
    "L-Isoleucine": 3,
    "L-Leucine": 4,
    "L-Methionine": 3,
    "L-Phenylalanine": 3,
    "L-Proline": 2,
    "L-Rhamnose": 3,
    "L-Tartaric acid": 4,
    "L-Tryptophane": 3,
    "L-Tyrosine": 3,
    "Malonic acid": 3,
    "Nicotinic acid": 2,
    "Octanoic acid": 4,
    "Pimelic acid": 4,
    "Raffinose": 4,
    "Succinic acid": 3,
    "Tetradecanoic acid": 6,
    "Vanillic acid": 4,
    # mixtures
    "Ia_01": 7,
    "Ia_02": 7,
    "Ia_03": 8,
    "Ia_04": 7,
    "Ib_01": 4,
    "Ic_01": 5,
    "IIa_01": 7,
    "IIa_03": 7,
    "IIa_04": 7,
    "IIb_01": 5,
    "IIc_01": 3,
    "IIIa_01": 7,
    "IIIb_01": 7,
    "IIIc_01": 5,
}


# Total intensity weights computed from single compound raster
# spectra in "exp" at noiseThreshold = 5*noiseStd
OLDBCompoundWeights = {
    "(E)-Cinnamic acid": 21859012.496749796,
    "1,2,3-Benzenetricarboxylic acid": 22211186.877651516,
    "1,2,4-Benzenetricarboxylic acid": 28527385.839007806,
    "4-Hydroxybenzoic acid": 17132620.513385423,
    "4-Hydroxycinnamic acid": 22403039.00800274,
    "4-Phenylbutanoic acid": 36354978.7523742,
    "9-Anthracenecarboxylic acid": 35001172.65861179,
    "Anthranilic acid": 36086178.92969638,
    "Benzoic acid": 29635603.682999723,
    "Biotin": 56740359.20936215,
    "Citric acid": 36143796.919918396,
    "D-Glucose": 97412798.24552464,
    "D-Mannose": 48325549.56628264,
    "D-Xylose": 81751109.9075893,
    "L-Cysteine": 26482232.482687145,
    "L-Glutamic acid": 37998891.80201849,
    "L-Glutamine": 44890563.45445999,
    "L-Isoleucine": 33853864.25405648,
    "L-Leucine": 18690289.75304223,
    "L-Methionine": 26495591.527344096,
    "L-Phenylalanine": 35685236.67435031,
    "L-Proline": 52887570.22402762,
    "L-Rhamnose": 47923534.29768606,
    "L-Tartaric acid": 6971518.753678279,
    "L-Tryptophane": 39452109.81456034,
    "L-Tyrosine": 35627413.78334053,
    "Malonic acid": 9633356.927163415,
    "Nicotinic acid": 19466861.819191758,
    "Octanoic acid": 22403034.708915677,
    "Pimelic acid": 20947561.12618703,
    "Raffinose": 63519645.654627636,
    "Succinic acid": 10103609.511354258,
    "Tetradecanoic acid": 10857794.299946425,
    "Vanillic acid": 15317233.912271747,
}

# Individual compound spectra were recorded at 3 mM. This yields expected concentration factors
# for each compound in the recombination dependent on the mix. These are recorded here.
mix4 = ["D-Glucose", "Pimelic acid", "1,2,3-Benzenetricarboxylic acid", "L-Tyrosine"]
expectedConcentrationFactors = {
    "Ia_01": dict([(c, 1.0) for c in OLDBCompoundWeights]),
    "Ib_01": dict([(c, 0.1) for c in OLDBCompoundWeights]),
    "Ic_01": dict([(c, 0.01) for c in OLDBCompoundWeights]),
    "IIa_01": dict([(c, 10.0) if c in mix4 else (c, 0.0) for c in OLDBCompoundWeights]),
    "IIb_01": dict([(c, 1.0) if c in mix4 else (c, 0.0) for c in OLDBCompoundWeights]),
    "IIc_01": dict([(c, 0.1) if c in mix4 else (c, 0.0) for c in OLDBCompoundWeights]),
    "IIIa_01": dict(
        [(c, 5.05) if c in mix4 else (c, 0.05) for c in OLDBCompoundWeights]
    ),
    "IIIb_01": dict(
        [(c, 0.55) if c in mix4 else (c, 0.05) for c in OLDBCompoundWeights]
    ),
    "IIIc_01": dict([(c, 0.1) if c in mix4 else (c, 0.05) for c in OLDBCompoundWeights]),
}






from scipy.ndimage import gaussian_filter
from shapely.geometry import Polygon
import numpy as np
from mcfnmr.utils.geometry import RectRegion
from mcfnmr.utils.parsesvg import loadRegions


class RasterSpectrum(object):
    """
    This is a class to handle 2D raster spectrum data.
    """

    def __init__(self, fullData, header, regionDefs=None, denoising_pars={}, name=None):
        self.header = header
        self.fullData = fullData
        self.pointDimensions = fullData.shape
        self.FRanges = header["FRanges"]
        self.smooth = denoising_pars.get("smoothRadius", 0.0)
        self.name = name

        # print("Loaded spectrum data for '%s'"%self.substance)
        print("Original grid dimension: %s" % str(fullData.shape))

        self.pointArea = self.calculatePointArea()
        # Raster points represent bins from here on
        self.fullData *= self.pointArea

        # Defining regions allows to disregard polygonal parts of the spectrum
        # as unknown. This allows to exclude noise, e.g.
        if regionDefs is not None:
            self.regions = loadRegions(**regionDefs)
        else:
            self.regions = self.getInnerRegion()

        # Range to use for noise measurement (in relative extent to fullspectral data raster), see getNoiseStd()
        self.noiseRegionRangeF1 = [0, 0.25]
        self.noiseRegionRangeF2 = [0.75, 1.0]
        # filter noise and smooth (sets self.origNoiseStd and self.noiseStd)
        self.noiseFilteringAndSmoothing(**denoising_pars)

        self.scanNr = header.get("ExpCount", None)
        self.smoothingHistory = []

    def noiseFilteringAndSmoothing(self, noiseFilterLevel=None, smoothRadius=0.0):
        # Calculating noiseStd before smoothing
        self.noiseStd = self.getNoiseStd()
        self.origNoiseStd = self.noiseStd
        print("Noise Std before processing: %g" % self.noiseStd)
        sm = smoothRadius if smoothRadius > 0 else self.smooth
        if sm > 0.0:
            self.smoothData(sm)
        self.noiseStd = self.getNoiseStd()
        print("Noise Std after processing: %g" % self.noiseStd)

        s2n = np.max(self.fullData) / self.noiseStd
        if noiseFilterLevel is not None:
            origMass = self.getTotalAveragePointIntensity(positiveWeightOnly=True)
            # Noise filter as multiples of self.noiseStd
            self.deleteNoise(noiseFilterLevel)

        mass = self.getTotalAveragePointIntensity()
        print("   Total avg point intensity: %g" % mass)
        print("   Max point intensity: %g" % np.max(self.fullData))
        if noiseFilterLevel is not None:
            print(
                "   Noise filter level: %d (corresponds to minimal signal intensity %g)"
                % (noiseFilterLevel, self.origNoiseStd * noiseFilterLevel)
            )
            print(
                "   Total cleared signal by noise filter: %g%s"
                % (100 * (1 - mass / origMass), "%")
            )
            print("    (Original noise std: %g)" % self.origNoiseStd)
            print("    (Original signal to noise: %g)" % s2n)
        else:
            print("   Noise std: %g" % self.noiseStd)
            print("   Signal to noise: %g" % s2n)
        print("")

    def isPointSpectrum(self):
        return False

    def getInnerRegion(self):
        ranges = self.FRanges
        coords = (
            (ranges[0][0], ranges[1][0]),
            (ranges[0][0], ranges[1][1]),
            (ranges[0][1], ranges[1][1]),
            (ranges[0][1], ranges[1][0]),
            (ranges[0][0], ranges[1][0]),
        )
        inner = Polygon(coords)
        return [inner]

    def getOuterRegions(self):
        # Construct regions covering the outside of the given ranges.
        # This is the minimal unknown region for the spectrum
        ranges = self.FRanges
        right = RectRegion((ranges[0][1], np.inf), (-np.inf, np.inf))
        left = RectRegion((-np.inf, ranges[0][0]), (-np.inf, np.inf))
        top = RectRegion((-np.inf, np.inf), (ranges[1][1], np.inf))
        bottom = RectRegion((-np.inf, np.inf), (-np.inf, ranges[1][0]))
        regions = [right, left, top, bottom]
        return regions

    def calculatePointArea(self):
        ranges, dimensions = self.FRanges, self.pointDimensions
        pointArea = (
            (ranges[0][-1] - ranges[0][0])
            * (ranges[1][-1] - ranges[1][0])
            / (dimensions[0] * dimensions[1])
        )
        return pointArea

    def smoothData(self, smoothRadius):
        # Apply a smoothing to the matrix
        print("Smoothing data with radius %g." % smoothRadius)
        self.fullData = gaussian_filter(
            self.fullData, sigma=smoothRadius, mode="mirror"
        )

    def getNoiseRegionData(self):
        ix11 = int(self.pointDimensions[0] * self.noiseRegionRangeF2[0])
        ix12 = int(self.pointDimensions[0] * self.noiseRegionRangeF2[1])
        ix21 = int(self.pointDimensions[1] * self.noiseRegionRangeF1[0])
        ix22 = int(self.pointDimensions[1] * self.noiseRegionRangeF1[1])
        return self.fullData[ix11:ix12, ix21:ix22].flatten()

    def getNoiseStd(self, useNegative=True):
        # Sample noise from corner of the spectrum
        noise = self.getNoiseRegionData()
        # Use negative part to estimate the std (avoids artifacts from peaks in the region,
        # although the selected region shouldn't hold any)
        if useNegative:
            noise = noise[noise <= 0]
            noise = np.concatenate((noise, -noise))
        noise_std = np.std(noise)
        return noise_std

    def deleteNoise(self, level):
        if self.fullData is None:
            print("Error at deleteNoise(): No data loaded.")
        else:
            assert level is not None
            self.fullData[self.fullData <= level * self.noiseStd] = 0
            self.noiseFiltered = True
            self.noiseStd = None

    def getRasterDataAsPeaks(
        self, nbin=(1, 1), xVar="F1", signalThreshold=None, cutGridToBin=True
    ):
        """
        This returns peak-coords and corresponding weights for the full grid intensity data without
        peak extraction. Returned coords are always the centers of the rasterpoints
        nbin - side length in raster points for binning
        signalThreshold - threshold to account for peaks relative to noise-std of the spectrum
        cutGridToBin - discard raster points if binning multiple does not match original raster point dimension
        """
        assert min(nbin) >= 1
        print(
            "Converting raster to peaks, nbin=%s, signal-to-noise threshold=%s)"
            % (nbin, str(signalThreshold))
        )

        assert self.FRanges[0][1] > self.FRanges[0][0]
        assert self.FRanges[1][1] > self.FRanges[1][0]

        nCutoff = (self.pointDimensions[0] % nbin[0], self.pointDimensions[1] % nbin[1])
        if not cutGridToBin:
            assert nCutoff[0] == 0
            assert nCutoff[1] == 0
            FRanges = self.FRanges
            ptDim = self.pointDimensions
            fullData = self.fullData
        else:
            # Grid point distances in full grid
            dF1 = (self.FRanges[0][1] - self.FRanges[0][0]) / self.pointDimensions[0]
            dF2 = (self.FRanges[1][1] - self.FRanges[1][0]) / self.pointDimensions[1]
            # Truncated dimensions (dividable by nbin)
            ptDim = [
                self.pointDimensions[0] - nCutoff[0],
                self.pointDimensions[1] - nCutoff[1],
            ]
            dCut = [nCutoff[0] * dF1, nCutoff[1] * dF2]
            FRanges = [list(self.FRanges[0]), list(self.FRanges[1])]
            fullData = self.fullData
            if FRanges[0][0] + dCut[0] <= 0.0:
                # cut left side of spectrum
                FRanges[0][0] += dCut[0]
                fullData = fullData[nCutoff[0] :, :]
            elif nCutoff[0] > 0:
                # cut right side of spectrum
                FRanges[0][1] -= dCut[0]
                fullData = fullData[: -nCutoff[0], :]
            if FRanges[1][0] + dCut[1] <= 0.0:
                # cut lower side of spectrum
                FRanges[1][0] += dCut[1]
                fullData = fullData[:, nCutoff[1] :]
            elif nCutoff[1] > 0:
                # cut upper side of spectrum
                FRanges[1][1] -= dCut[1]
                fullData = fullData[:, : -nCutoff[1]]

        if signalThreshold is not None:
            if self.noiseStd is None:
                noiseStd = self.getNoiseStd()
            else:
                noiseStd = self.noiseStd
            noiseLevel = noiseStd * signalThreshold
            print(
                "signal-to-noise threshold %g corresponds to intensity %g"
                % (signalThreshold, noiseLevel)
            )
            totalCleared = np.sum(fullData[fullData < noiseLevel])
            print("Clearing total intensity of %g" % (totalCleared))
            fullData[fullData < noiseLevel] = 0.0

        # Grid point distances if pointDimension % nbin == 0
        dF1 = (FRanges[0][1] - FRanges[0][0]) * nbin[0] / ptDim[0]
        dF2 = (FRanges[1][1] - FRanges[1][0]) * nbin[1] / ptDim[1]

        F10 = FRanges[0][0] + 0.5 * dF1
        F20 = FRanges[1][0] + 0.5 * dF2

        gridShape = (int(ptDim[0] / nbin[0]), int(ptDim[1] / nbin[1]))
        integralWeights = np.zeros(gridShape)
        integralCoords = np.zeros(list(gridShape) + [2])
        pointClusterArea = self.pointArea * nbin[0] * nbin[1]

        for i in range(gridShape[0]):
            for j in range(gridShape[1]):
                pointClusterMeanIntensity = (
                    np.mean(
                        fullData[
                            i * nbin[0] : (i + 1) * nbin[0],
                            j * nbin[1] : (j + 1) * nbin[1],
                        ]
                    )
                    / self.pointArea
                )
                integralWeights[i, j] = pointClusterMeanIntensity * pointClusterArea
                # # TODO: replace the above by equivalent (will cause numerical precision errors):
                # integralWeights[i, j] = np.sum(
                #         fullData[
                #             i * nbin[0] : (i + 1) * nbin[0],
                #             j * nbin[1] : (j + 1) * nbin[1],
                #         ]
                #     )
                if xVar == "F1":
                    integralCoords[i, j, :] = F10 + i * dF1, F20 + j * dF2
                else:
                    integralCoords[i, j, :] = F20 + j * dF2, F10 + i * dF1

        npts = integralWeights.size
        coords, weights = integralCoords.reshape((npts, 2)), integralWeights.reshape(
            (npts,)
        )
        print("   (original raster dimensions: %s)" % str(self.pointDimensions))
        print("   (binned grid dimensions: %s)" % str(gridShape))
        return coords, weights, pointClusterArea

    def getTotalAveragePointIntensity(
        self, positiveRangeOnly=False, positiveWeightOnly=False
    ):
        loaded = self.fullData is not None
        if not loaded:
            self.load()

        if positiveWeightOnly:
            data = self.fullData
            data[data < 0] = 0
        else:
            data = self.fullData
        if positiveRangeOnly:
            xspan = np.linspace(
                self.FRanges[1][0], self.FRanges[1][1], self.pointDimensions[1]
            )
            yspan = np.linspace(
                self.FRanges[0][0], self.FRanges[0][1], self.pointDimensions[0]
            )
            xiPos, yiPos = np.argmax(xspan > 0), np.argmax(yspan > 0)
            res = np.sum(data[yiPos:, xiPos:]) / (len(xspan) * len(yspan))
        else:
            res = np.sum(data)

        if not loaded:
            self.unload()
        return res


def sampleBlankSpectrum(rasterSpec, seed):
    # Use noise region of rasterSpec to generate a blank Spectrum
    # With the same noise distribution
    noise = rasterSpec.getNoiseRegionData().flatten()
    dims = rasterSpec.pointDimensions
    rng = np.random.default_rng(seed)
    fullData = rng.choice(noise.flatten(), replace=True, size=dims)
    blank = RasterSpectrum(
        fullData,
        header=rasterSpec.header,
        regionDefs=None,
        name=rasterSpec.name + "(blank, seed%d)" % seed,
    )
    blank.regions = rasterSpec.regions
    return blank



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from copy import deepcopy

from mcfnmr.demodata import caseSpecC, caseSpecB, caseSpecA, caseSpecD
from mcfnmr.demodata.loading import (
    loadMMMixtureLists,
    loadMMMixPointSpectra,
    loadMMLibrary,
    loadSMARTMINERBenchmarks,
)
from mcfnmr.routines.classification import classificationStats
from mcfnmr.routines.utils import (
    cache_dir,
    runMetaboMinerFits,
    prepareParameters,
    figs_dir,
    totalWeights,
    makeResultsDF,
    makeTargetSpecification,
    collect_and_serialize_MM_grid_spectra,
)
from mcfnmr.demodata import mix_selections
from mcfnmr.config import DEBUG


def plotMetaboMinerComparison(
    libID,
    df,
    plotOnlyFScore=False,
    title=None,
    show=True,
    show_legend=True,
    plot_width=2.5,
    plot_height=None,
    pal=None,
):
    if libID is not None:
        df = df[df["libID"] == libID]
    if title is None:
        title = f"Comparison MetaboMiner-EMD lib '{libID}'"
    if pal is None:
        pal = sns.color_palette("magma", n_colors=7)
    nrows = 1 if plotOnlyFScore else 3
    plt.style.use("bmh")
    if plot_height:
        figsize = (plot_width, plot_height)
    else:
        figsize = (plot_width, 2 * nrows)
    fig, axes = plt.subplots(
        figsize=figsize, nrows=nrows, sharex=True, layout="constrained"
    )
    if plotOnlyFScore:
        axes = [axes]
    variables = ("F1 score", "precision", "recall")
    fig.suptitle(title)
    for ax, var in zip(axes, variables):
        ax.set_xlim((0.0, 1.0))
        sns.barplot(df, y="mixID", x=var, hue="method", palette=pal, ax=ax)

    if DEBUG > 0:
        dfg = df.groupby("method")
        for method in dfg.groups:
            print("Method:", method)
            g = dfg.get_group(method)
            print("mixes:", list(g["mixID"]))
            print("avg. fscore:", np.mean(g["F1 score"]))
            print("avg. recall:", np.mean(g["recall"]))
            print("avg. precision:", np.mean(g["precision"]))
            gg = g.groupby("mixID")
            for mix in gg.groups:
                dfgg = gg.get_group(mix)
                print("   mixID:", mix)
                print("     fscore:", list(dfgg["F1 score"]))
                print("     recall:", list(dfgg["recall"]))
                print("     precision:", list(dfgg["precision"]))
                if mix == "all":
                    nFN = nFP = nTP = None
                else:
                    nTP = dfgg['TP'].iloc[0]
                    nTP = len(nTP) if type(nTP) is list else nTP
                    nFP = dfgg['FP'].iloc[0]
                    nFP = len(nFP) if type(nFP) is list else nFP
                    nFN = dfgg['FN'].iloc[0]
                    nFN = len(nFN) if type(nFN) is list else nFN
                print(f"     TP ({nTP}):", list(dfgg["TP"]))
                print(f"     FN ({nFN}):", list(dfgg["FN"]))
                print(f"     FP ({nFP}):", list(dfgg["FP"]))

    if not show_legend:
        ax.legend(loc=2).remove()
    fn = figs_dir / f"{title}_fscores.svg"
    fig.savefig(fn)
    print(f"\nSaved figure '{fn}'")

    if show:
        plt.show()
    else:
        plt.close(fig)


def makeMethodName(caseSpec):
    name = "MCF (%s)" % caseSpec["name"]
    return name


def evaluateMCFPerformance(pars, caseSpec, assignmentRadius, th, load=True):
    pars["isolated_fit"] = caseSpec["isolated_fit"]
    pars["assignmentRadius"] = assignmentRadius
    pars["libID"] = caseSpec["libID"]

    mixIDs = caseSpec["mixIDs"]
    mixSubs = loadMMMixtureLists()
    lib = loadMMLibrary(pars["libID"])

    if pars["pointTarget"]:
        mixSpecs = loadMMMixPointSpectra()
    else:
        pars["binning"] = None
        mixSpecs = collect_and_serialize_MM_grid_spectra(mixIDs, pars)

    results = runMetaboMinerFits(
        mixIDs=mixIDs, base_pars=pars, lib=lib, mixSpecs=mixSpecs, load=load
    )

    dfs = {}
    for mixID in mixIDs:
        result = results[mixID]
        if pars["pointTarget"]:
            pars["targetSpecID"] = mixID + "(peaklist)"
        else:
            pars["targetSpecID"] = mixID
            pars["specificTargetSpecID"] = makeTargetSpecification(
                pars, pars["targetSpecID"], mixID
            )
        dfTargets, dfCompounds = makeResultsDF(
            [{"res": result, "pars": pars}], originalWeights=totalWeights(lib)
        )
        dfs[mixID] = {
            "targets": dfTargets,
            "compounds": dfCompounds,
            "targetSpecID": pars["targetSpecID"],
            "targetTotalWeight": result.originalWeightY,
        }
    scores = classificationStats(dfs, pars, th, mixSubs, mixIDs)

    df = {
        "mixID": [],
        "libID": [],
        "assignmentRadius": [],
        "binning": [],
        "absorptionCost": [],
        "precision": [],
        "recall": [],
        "F1 score": [],
        "cutoffFlag": [],
        "hueID": [],
        "method": [],
        "FP": [],
        "TP": [],
        "FN": [],
    }

    method = makeMethodName(caseSpec)
    for mixID, data in scores.items():
        df["mixID"].append(mixID)
        df["precision"].append(data["classification"]["precision"])
        df["recall"].append(data["classification"]["recall"])
        df["F1 score"].append(data["classification"]["fscore"])
        df["FP"].append(data["classification"]["FP"])
        df["TP"].append(data["classification"]["TP"])
        df["FN"].append(data["classification"]["FN"])
        df["libID"].append(caseSpec["libID"])
        df["assignmentRadius"].append(assignmentRadius)
        df["absorptionCost"].append(pars["absorptionCost"])
        df["binning"].append(pars["binning"])
        df["method"].append(method)

    for k in list(df.keys()):
        if len(df[k]) == 0:
            df[k] = None
    df = pd.DataFrame(df)
    return df


def main(cfg, show=False):
    wdir = cache_dir / "method_comparison"
    assert cfg["task"] == "comparison_classification"
    assert cfg["lib"] == "MetaboMiner"
    assert cfg["mix"] in ["N925", "N987", "N988", "all"]

    kwargs = dict(
        normalizeY=False,
        specDomain="MetaboMiner",
        incrementalFit=False,
        noiseFilterLevel=None,
        smoothing=0,
        workingDir=cfg["outdir"],
        recompute=cfg.get("recompute", False),
    )
    pars = prepareParameters(targetSpecID=cfg["mix"], libID=cfg["lib"], **kwargs)

    if cfg["mix"] == "all":
        mixIDs = mix_selections["SMART-Miner"]
    else:
        mixIDs = [cfg["mix"]]

    # assignment radii (hard-coded for incremental approach)
    arSpan = [0.02, 0.03, 0.04, 0.05, 0.06, 0.08, 0.1, 0.12, 0.15, 0.175, 0.2, 0.25]

    includeAverageOverAllMixes = True
    # Select mix selection, assignment radius and detection threshold
    mixSelection = "SMART-Miner"

    # Hard-coded optimal assignment radii and detection thresholds for comparison
    optPars = {
        "setupA": {"ar": 0.06, "th": 0.003},
        "setupB": {"ar": 0.05, "th": 0.003},
        "setupC": {"ar": 0.08, "th": 0.003},
        "setupD": {"ar": 0.06, "th": 0.005},
    }

    caseSpecIncr = deepcopy(caseSpecC)
    testsets = [
        {"case": caseSpecA, **optPars["setupA"]},  # "standard"
        {"case": caseSpecB, **optPars["setupB"]},  # "independent"
        {"case": caseSpecIncr, **optPars["setupC"]},  # "incremental"
        {"case": caseSpecD, **optPars["setupD"]},  # "peak list"
    ]

    libID = "Biofluid ( all )"
    for v in testsets:
        v["case"]["libID"] = libID

    # For incremental fit, prepare previousARSteps to allow loading
    # previously computed incremental fit result
    # → set load=True for evaluateMCFPerformance()
    previousARSteps = [x for x in arSpan if x < optPars["setupC"]["ar"]]
    previousAssignments = dict()
    for mixID in caseSpecIncr["mixIDs"]:
        previousAssignments[mixID] = dict(previousARSteps=deepcopy(previousARSteps))
    caseSpecIncr["previousAssignments"] = previousAssignments

    # Generate data for selected mixes
    dfMCFall = []
    for d in testsets:
        print("\n#  ", d["case"]["name"])
        arSpan_restricted = [float(ar) for ar in arSpan if ar <= d["ar"]]
        assert d["ar"] == arSpan_restricted[-1]
        kwargs = dict(
            arSpan=arSpan_restricted,
            incrementalFit=d["case"]["incrementalFit"],
            isolated_fit=d["case"]["isolated_fit"],
            noiseFilterLevel=None,
            pointTarget=d["case"]["pointTarget"],
            smoothing=0,
            workingDir=wdir,
            recompute=False,
        )
        pars = prepareParameters(
            specDomain="metabominer", targetSpecID=None, libID=None, **kwargs
        )
        dfMCFall.append(
            evaluateMCFPerformance(
                pars, d["case"], assignmentRadius=d["ar"], th=d["th"], load=True
            )
        )
    dfSM = loadSMARTMINERBenchmarks()
    dfAll = pd.concat(dfMCFall + [dfSM])

    # Filter for selected mixes
    ix = [
        (mid in mixIDs) and lid == libID
        for mid, lid in zip(dfAll["mixID"], dfAll["libID"])
    ]
    dfAll = dfAll.iloc[ix]

    if includeAverageOverAllMixes:
        # include the average F1 (and other numeric columns) over all mixes in dfAll
        numeric_cols = ["precision", "recall", "F1 score"]
        dfavg = {c: [] for c in dfAll.columns}
        dfg = dfAll.groupby("method")
        for method in dfg.groups:
            for k in dfAll.columns:
                if k == "method":
                    dfavg[k].append(method)
                elif k in numeric_cols:
                    dfm = dfg.get_group(method)
                    dfavg[k].append(np.mean(dfm[k]))
                elif k == "mixID":
                    dfavg[k].append("all")
                elif k == "libID":
                    dfavg[k].append(libID)
                else:
                    dfavg[k].append(None)

        dfavg = pd.DataFrame(dfavg)
        dfAll = pd.concat([dfAll, dfavg])

    # Compare full library against other algorithms
    ars = ["%g" % t["ar"] for t in testsets]
    ths = ["%g" % t["th"] for t in testsets]

    title = (
        f"Comparison MetaboMiner-MCF with lib '{libID}', mixes: '{mixSelection}', ar: %s, th: %s"
        % (ths, ars)
    )
    plotMetaboMinerComparison(
        libID=libID, df=dfAll, title=title, plotOnlyFScore=True, show=show
    )


if __name__ == "__main__":
    # Test run
    cfg = dict(
        lib="MetaboMiner",
        mix="all",
        recompute=False,
        task="comparison_classification",
        ar=dict(
            A=0.06,
            B=0.05,
            C=0.08,
            D=0.06,
        ),
        th=dict(
            A=0.003,
            B=0.003,
            C=0.003,
            D=0.005,
        ),
        outdir=cache_dir,
    )

    main(cfg, show=True)



import os
import numpy as np
import svglib.svglib as svg
import reportlab.graphics
from shapely.geometry import (
    Polygon,
    MultiPolygon,
)


def getAllShapes(contentList, lvl=0, verb=False):
    allShapes = []
    for c in contentList:
        if type(c) == reportlab.graphics.shapes.Group:
            shapes = getAllShapes(c.getContents(), lvl + 1, verb=verb)
            allShapes.extend(shapes)
        elif type(c) == reportlab.graphics.shapes.Path:
            if verb:
                print("Found Path ('%s') at depth %d." % (str(c), lvl))
            allShapes.append(c)
        elif type(c) == reportlab.graphics.shapes.Rect:
            if verb:
                print("Found Rect ('%s') at depth %d." % (str(c), lvl))
            allShapes.append(c)
        else:
            if verb:
                print("Found entry of type '%s' at depth %d." % (type(c), lvl))
    return allShapes


# drawing operators (probably used in reportlab)
operatorIDs = {"START": 0, "CONNECT": 1, "CLOSE": 3}


def makePolys(shapes, verb=0):
    polys = []
    for i, p in enumerate(shapes):
        if type(p) == reportlab.graphics.shapes.Rect:
            x, y, w, h = p.x, p.y, p.width, p.height
            pts = [(x, y), (x, y + h), (x + w, y + h), (x + w, y)]
        elif type(p) == reportlab.graphics.shapes.Path:
            pts, ops = p.points, p.operators
            if ops[-1] != operatorIDs["CLOSE"]:
                # No closed segment
                if verb > 0:
                    print("Path %d (%s) is not closed." % (i, p))
                continue
            xCoords = np.array(pts)[0::2]
            yCoords = np.array(pts)[1::2]
            pts = list(zip(xCoords, yCoords))
            if len(pts) < 3:
                if verb > 0:
                    print("Path %d (%s) only consists of %d points." % (i, p, len(pts)))
                continue
        if verb > 1:
            print("Adding polygon for shape %d (%s)" % (i, p))
        poly = Polygon(pts)
        polys.append(poly)
    if verb > 0:
        print("Generated %d polygons" % len(polys))
    return polys


def separateReferenceRect(polys):
    # Find poly with largest area to identify full area range
    areas = [p.area for p in polys]
    boundsIx = np.argmax(areas)
    bounds = polys[boundsIx]
    coords = bounds.exterior.coords
    # Test if fullRange is a rectangle
    errmsg = "Reference area must be \n   (1) largest area, \n   (2) rectangular"
    if len(coords) != 5:
        print(
            "Rectangle bounding area must have 4 points (+1 repeating the first to close it in shapely)"
        )
        raise Exception(errmsg)
    xcoords, ycoords = coords.xy
    minx, maxx, miny, maxy = min(xcoords), max(xcoords), min(ycoords), max(ycoords)
    boundsBB = bounds.bounds
    if (
        (boundsBB != (minx, miny, maxx, maxy))
        or (len(set(xcoords)) != 2)
        or (len(set(ycoords)) != 2)
    ):
        print(
            "Rectangle bounding area must not be rotated and rectangular (equal to its bounding box)"
        )
        raise Exception(errmsg)
    return bounds, [p for i, p in enumerate(polys) if i != boundsIx]


def normalizeCoords(polys, bounds):
    # Transform everything to the range [0,1]x[0,1]
    bb = bounds.bounds
    x0, y0 = bb[0], bb[1]
    scalex, scaley = bb[2] - bb[0], bb[3] - bb[1]
    transform = lambda x, y: ((x - x0) / scalex, (y - y0) / scaley)
    transformedPolys = []
    for p in polys:
        xcoords, ycoords = p.exterior.coords.xy
        transformedPts = [transform(x, y) for x, y in zip(xcoords, ycoords)]
        transformedPoly = Polygon(transformedPts)
        transformedPolys.append(transformedPoly)
    transformedBounds = Polygon([(0, 0), (0, 1), (1, 1), (1, 0)])
    return transformedPolys, transformedBounds


def scaleToRange(polys, ranges):
    # This expects polys in a normalized range:
    for p in polys:
        assert max(p.bounds) <= 1
        assert min(p.bounds) >= 0
    # Whether x and y regions are inverted
    invertX = ranges["x"][0] > ranges["x"][1]
    invertY = ranges["y"][0] < ranges["y"][1]  # svg has growing coords to bottom
    x0, x1 = min(ranges["x"]), max(ranges["x"])
    y0, y1 = min(ranges["y"]), max(ranges["y"])
    rx, ry = x1 - x0, y1 - y0

    if invertX:
        transformX = lambda x: x0 + rx * (1.0 - x)
    else:
        transformX = lambda x: x0 + rx * x
    if invertY:
        transformY = lambda y: y0 + ry * (1.0 - y)
    else:
        transformY = lambda y: y0 + ry * y

    transform = lambda xy: (transformX(xy[0]), transformY(xy[1]))

    transformed = []
    for i, p in enumerate(polys):
        shell, holes = [], []
        xCoords, yCoords = p.exterior.coords.xy
        for pt in zip(xCoords, yCoords):
            shell.append(transform(pt))
        for pp in p.interiors:
            hole = []
            xCoords, yCoords = pp.xy
            for pt in zip(xCoords, yCoords):
                hole.append(transform(pt))
            holes.append(hole)
        transformed.append(Polygon(shell, holes))
    return transformed


def invertPolys(polys, bounds, verb=False):
    # Take the complement of each poly within bounds
    inverted = []
    for i, p in enumerate(polys):
        if verb:
            print("Polygon %d:\n%s" % (i, p))
        complement = bounds.difference(p)
        if verb:
            print(
                "Complement of %s %d is %s"
                % (
                    str(type(p)).split(".")[-1][:-2],
                    i,
                    str(type(complement)).split(".")[-1][:-2],
                )
            )
        if type(complement) == MultiPolygon:
            geoms = complement.geoms
            for j, pp in enumerate(geoms):
                if verb:
                    print("   complement part %d.exterior: %s" % (j, pp.exterior))
                    print(
                        "   complement part %d.interiors: %s" % (j, list(pp.interiors))
                    )
                assert type(pp) == Polygon
                inverted.append(pp)
        elif type(complement) == Polygon:
            inverted.append(complement)
            if verb:
                print("  complement.exterior:", complement.exterior)
                print("  complement.interiors:", list(complement.interiors))
        else:
            print("Unexpected type of complement!")
            raise Exception("Unexpected type of complement: %s" % type(complement))
    return inverted


def mergePolys(polys, verb=False):
    mergedPolys = []
    for i, poly in enumerate(polys):
        # NOTE: In the following, we assume that no higher
        # nesting of polygons, i.e., only single holes occur
        # Further we assume that any holes do only pierce one
        # containing polygon
        containment = False
        for j in range(len(mergedPolys)):
            p = mergedPolys[j]
            if p.contains(poly):
                if verb:
                    print(
                        "Polygon %d is contained in merged polygon %d (extend of smaller: %s)"
                        % (i, j, poly.bounds)
                    )
                assert len(poly.interiors) == 0
                mergedPolys[j] = Polygon(
                    p.exterior, holes=list(p.interiors) + [poly.exterior]
                )
                containment = True
                break
            elif p.within(poly):
                if verb:
                    print(
                        "Polygon %d contains merged polygon %d (extend of smaller: %s)"
                        % (i, j, p.bounds)
                    )
                assert len(p.interiors) == 0
                mergedPolys[j] = Polygon(poly.exterior, holes=[p.exterior])
                containment = True
                break
        if not containment:
            # No containment relation with any existing
            if verb:
                print(
                    "New polygon %d disjoint from all previous. Adding it to collection."
                    % (i)
                )
            mergedPolys.append(poly)

    # Test: to check validity (no intersections of elements), make collection
    MultiPolygon(mergedPolys)

    return mergedPolys


def cropAndDiscardUncontained(polys, bounds):
    contained = []
    for i, p in enumerate(polys):
        if not p.intersection(bounds):
            print(
                "Polygon %d doesn't intersect with bounding rectangle => discarding\n   (%s)"
                % (i, p)
            )
        else:
            cropped = p.intersection(bounds)
            assert type(cropped) == Polygon
            contained.append(cropped)
    return contained


def loadRegions(
    svgFile, originalRanges, invertRegions=False, flipAxes=False, verb=True
):
    if not os.path.exists(svgFile):
        return []
    drawing = svg.svg2rlg(svgFile)
    if drawing is None:
        print("Couldn't parse regions file '%s'" % svgFile)
    allShapes = getAllShapes(drawing.getContents())
    polys = makePolys(allShapes)
    bounds, polys = separateReferenceRect(polys)
    polys, bounds = normalizeCoords(polys, bounds)
    polys = cropAndDiscardUncontained(polys, bounds)
    polys = mergePolys(polys)
    if invertRegions:
        polys = invertPolys(polys, bounds)
    polys = scaleToRange(polys, originalRanges)
    if flipAxes:
        polys = swapAxes(polys)
    if verb:
        print("Loaded regions:")
        for i, p in enumerate(polys):
            print("   Polygon %d: %s" % (i, p))
    return polys


def swapAxes(polys):
    swapped = []
    for p in polys:
        newShell = list(zip(p.exterior.xy[1], p.exterior.xy[0]))
        newHoles = []
        for i in p.interiors:
            newHoles.append(list(zip((i.xy[1], i.xy[0]))))
        swapped.append(Polygon(shell=newShell, holes=newHoles))
    return swapped


def test1():
    ## Input
    # Source file for shapes
    svgfn = "/home/leo/Workspace/NMR_analysis/data/full spectra/DOM/HSQC_NELHA.007.001.2rr_processed.svg"
    # Whether to ignore the complements of the provided regions, rather
    invertRegions = True
    # Ranges of the underlying data (left-to-right on x-axis and bottom-to-top on y-axis)
    ranges = {"x": (9.5, -2.5), "y": (150, 0)}

    drawing = svg.svg2rlg(svgfn)
    contents = drawing.getContents()
    allShapes = getAllShapes(drawing.getContents(), 0)
    polys = makePolys(allShapes)
    bounds, polys = separateReferenceRect(polys)
    polys, bounds = normalizeCoords(polys, bounds)

    # Test cases:
    # (1) discard outside bounds
    polyOutsideBounds = Polygon([(1000, 1000), (1000, 1001), (1001, 1001)])
    # (2) handle poly which exceeds bounding box range
    # polyDisectingBounds = Polygon([(-0.1, -0.1), (1.1,1.1), (1.2,1.1)])
    polyDisectingBounds = Polygon([(0.4, -0.1), (0.4, 1.1), (0.6, 1.1), (0.6, -0.1)])
    # (3) handle poly which lies within another
    polyWithinOther = Polygon([(0.6, 0.1), (0.65, 0.2), (0.6, 0.2)])
    # Add test cases
    polys.extend([polyOutsideBounds, polyDisectingBounds, polyWithinOther])

    polys = cropAndDiscardUncontained(polys, bounds)
    polys = mergePolys(polys)
    if invertRegions:
        polys = invertPolys(polys, bounds)
    polys = scaleToRange(polys, ranges)

    for i, p in enumerate(polys):
        assert type(p) == Polygon
        print("Polygon %d: %s" % (i, p))


def test2():
    # Source file for shapes
    svgFile = "/home/leo/Workspace/NMR_analysis/data/full spectra/DOM/HSQC_NELHA.007.001.2rr_processed.svg"
    # Whether to ignore the complements of the provided regions, rather
    invertRegions = False
    # Ranges of the underlying data (left-to-right on x-axis and bottom-to-top on y-axis)
    originalRanges = {"x": (9.5, -2.5), "y": (150, 0)}

    loadRegions(svgFile, originalRanges, invertRegions)


if __name__ == "__main__":
    test2()



#! /usr/bin/python

import sys
from .oldbdata import prepareOLDBdata
from .metabominer import prepareMM
from .bmrb import prepareBMRB

def main():
    if len(sys.argv) == 1:
        print("Please specify, which dataset to download. ('bmrb', 'oldb', 'metabominer', 'all')")
        sys.exit(1)
    elif "all" in sys.argv:
        request = ["oldb", "bmrb", "metabominer"]
    else:
        request = sys.argv[1:]
    if "oldb" in request:
        prepareOLDBdata()
    if "metabominer" in request:
        prepareMM()
    if "bmrb" in request:
        prepareBMRB()



if __name__ == "__main__":
    main()


import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from pprint import pp
import pickle
from collections import defaultdict

from mcfnmr.demodata.loading import (
    loadMetaboMinerSpectrum,
    loadMMLibrary,
    loadMMMixtureLists,
    loadMMMixPointSpectra,
    buildHMDBID2NameMap,
    buildName2HMDBIDMap,
)
from mcfnmr.routines.utils import (
    prepareParameters,
    cache_dir,
    makeResultsDF,
    totalWeights,
    figs_dir,
    runMetaboMinerFits,
    collect_and_serialize_MM_grid_spectra,
)
from mcfnmr.utils.pointspectrum import makeRasterPointSpectrum, makeMix
from mcfnmr.demodata import case_dict
from mcfnmr.routines.classification import classificationStats
from mcfnmr.utils.plotting import plotLibOnTarget, plotRasterSpectrum, sqrt_transfo
from mcfnmr.demodata import binning, metaboMinerLibNames
from mcfnmr.routines.plotting import plotPerformanceScanResults, plotSlices
from mcfnmr.utils.rasterspectrum import sampleBlankSpectrum
from mcfnmr.routines.comparison_classification import plotMetaboMinerComparison
from mcfnmr.config import DATADIR, METABOMINER_DIR, SMARTMINER_DATADIR
from data.MetaboMiner.mixture_contents.map_to_HMDB import findHMDB_ID
from mcfnmr.utils.system import get_mcfnmr_home


outdir = figs_dir / "MM_N926"
if not outdir.exists():
    os.makedirs(outdir)
    print(f"Created directory '{outdir}'")


def makePerformanceTableCOLMAR(mixID, match_cutoff, peaks="deeppicker"):
    assert peaks in ["deeppicker", "metabominer"]

    COLMARDIR = DATADIR / "COLMAR"
    if peaks == "deeppicker":
        results = pd.read_csv(
            COLMARDIR / f"compound_report_{mixID}(dp)_0.04_COLMARm_refd.txt", sep=" "
        )
    else:
        results = pd.read_csv(
            COLMARDIR / f"compound_report_{mixID}_0.03_HSQC_refd.txt", sep=" "
        )

    mixSubs = loadMMMixtureLists()[mixID]

    df = {
        "libID": [],
        "F1 score": [],
        "precision": [],
        "recall": [],
        "method": [],
        "mixID": [],
        "ar_opt": [],
        "th_opt": [],
        "TP": [],
        "FN": [],
        "FP": [],
    }

    map_name2HMDB = buildName2HMDBIDMap(verb=0)
    all_names = sorted(map_name2HMDB.keys())
    detections = set()
    for name, match_ratio in zip(results["Name"], results["Matching_ratio"]):
        HMDB_ID = findHMDB_ID(name, map_name2HMDB, all_names)
        print(f"{name} → {HMDB_ID}")
        if HMDB_ID in detections:
            print(f"    → {HMDB_ID} ({name}) repeated.")
        if match_ratio >= match_cutoff:
            detections.add(HMDB_ID)
    df["libID"].append("COLMAR")
    if peaks == "deeppicker":
        df["method"].append(f"COLMARm (dp)")
    else:
        df["method"].append(f"COLMAR-HSQC")
    df["mixID"].append(mixID)
    df["ar_opt"].append(None)
    df["th_opt"].append(None)
    TP = detections.intersection(mixSubs)
    FP = detections.difference(mixSubs)
    FN = set(mixSubs).difference(detections)
    df["TP"].append(len(TP))
    df["FP"].append(len(FP))
    df["FN"].append(len(FN))
    recall = len(TP) / (len(TP) + len(FN))
    precision = len(TP) / (len(TP) + len(FP))
    fscore = (
        precision * recall * 2 / (precision + recall)
        if (precision + recall) > 0
        else 0.0
    )
    df["recall"].append(recall)
    df["precision"].append(precision)
    df["F1 score"].append(fscore)

    return pd.DataFrame(df), dict(TP=TP, FP=FP, FN=FN)


def makePerformanceTableMM(mixID, libID = "MetaboMiner-all"):
    table1 = pd.read_csv(
        METABOMINER_DIR / "tables_Xia_etal_2008" / "table1.txt", sep=" "
    )
    # table2 = pd.read_csv(METABOMINER_DIR / "tables_Xia_etal_2008" / "table2.txt", sep=" ")
    # table3 = pd.read_csv(METABOMINER_DIR / "tables_Xia_etal_2008" / "table3.txt", sep=" ")
    table_N907 = pd.read_csv(
        METABOMINER_DIR / "tables_Xia_etal_2008" / "table_N907.txt", sep=" "
    )
    
    table1 = pd.concat((table1, table_N907))
    table1.index = list(range(len(table1)))
    
    assert "N880" != mixID  # No performance provided for full lib

    df = {
        "libID": [],
        "F1 score": [],
        "precision": [],
        "recall": [],
        "method": [],
        "mixID": [],
        "ar_opt": [],
        "th_opt": [],
        "TP": [],
        "FN": [],
        "FP": [],
    }

    mix2sample = dict(N925="A", N926="D", N907="D(pH8)", N987="B", N988="C")
    sample = mix2sample[mixID]
    
    if libID == "MetaboMiner-all":
        dfMM = table1.loc[table1.groupby("Method").groups[libID]]
    elif libID == "MetaboMiner-plasma-common":
        dfMM = table1.loc[table1.groupby("Method").groups["MetaboMiner-sp"]]
        
    if sample in dfMM.groupby("Sample").groups:
        dfMM = dfMM.loc[dfMM.groupby("Sample").groups[sample]]
    else:
        # empty df
        dfMM = dfMM.iloc[0:0,:]

    df["libID"].append(libID)
    df["method"].append("MetaboMiner")
    df["ar_opt"].append(None)
    df["th_opt"].append(None)
    df["mixID"].append(mixID)
    df["F1 score"].append(dfMM["F-score"].iloc[0] / 100.0)
    df["precision"].append(dfMM["Precision"].iloc[0] / 100.0)
    df["recall"].append(dfMM["Recall"].iloc[0] / 100.0)
    df["TP"].append(dfMM["TP"].iloc[0])
    df["FN"].append(dfMM["FN"].iloc[0])
    df["FP"].append(dfMM["FP"].iloc[0])

    return pd.DataFrame(df)


def makePerformanceTableSMARTMiner(mixID):
    subs = pd.read_csv(
        SMARTMINER_DATADIR / f"predicted_{mixID}.txt", sep=";", header=None
    )
    mixSubs = loadMMMixtureLists()[mixID]

    df = {
        "libID": [],
        "F1 score": [],
        "precision": [],
        "recall": [],
        "method": [],
        "mixID": [],
        "ar_opt": [],
        "th_opt": [],
        "TP": [],
        "FN": [],
        "FP": [],
    }

    map_name2HMDB = buildName2HMDBIDMap(verb=0)
    all_names = sorted(map_name2HMDB.keys())
    detections = set()
    for name in subs.iloc[:, 0]:
        HMDB_ID = findHMDB_ID(name, map_name2HMDB, all_names)
        print(f"{name} → {HMDB_ID}")
        if HMDB_ID in detections:
            print(f"    → {HMDB_ID} ({name}) repeated.")
        detections.add(HMDB_ID)

    df["libID"].append("SMART-Miner")
    df["method"].append(f"SMART-Miner")
    df["mixID"].append(mixID)
    df["ar_opt"].append(None)
    df["th_opt"].append(None)
    TP = detections.intersection(mixSubs)
    FP = detections.difference(mixSubs)
    FN = set(mixSubs).difference(detections)
    df["TP"].append(len(TP))
    df["FP"].append(len(FP))
    df["FN"].append(len(FN))
    recall = len(TP) / (len(TP) + len(FN))
    precision = len(TP) / (len(TP) + len(FP))
    fscore = (
        precision * recall * 2 / (precision + recall)
        if (precision + recall) > 0
        else 0.0
    )
    df["recall"].append(recall)
    df["precision"].append(precision)
    df["F1 score"].append(fscore)

    return pd.DataFrame(df), dict(TP=TP, FP=FP, FN=FN)


def mergeOtherResults(df, colmar_match_cutoffs, libID_MM = "MetaboMiner-all"):
    classes = dict()
    df_sm_N926, classes["SMART-Miner"] = makePerformanceTableSMARTMiner("N926")
    df_sm_N907, classes["SMART-Miner"] = makePerformanceTableSMARTMiner("N907")
    df_colmar_N926_pl, classes["COLMAR"] = makePerformanceTableCOLMAR(
        "N926", match_cutoff=colmar_match_cutoffs["N926"], peaks="metabominer"
    )
    df_colmar_N926, classes["COLMAR"] = makePerformanceTableCOLMAR(
        "N926", match_cutoff=colmar_match_cutoffs["N926(dp)"]
    )
    df_colmar_N907_pl, classes["COLMAR"] = makePerformanceTableCOLMAR(
        "N907", match_cutoff=colmar_match_cutoffs["N907"], peaks="metabominer"
    )
    df_colmar_N907, classes["COLMAR"] = makePerformanceTableCOLMAR(
        "N907", match_cutoff=colmar_match_cutoffs["N907(dp)"]
    )
    df_MM = pd.concat((makePerformanceTableMM("N926", libID=libID_MM), 
                       makePerformanceTableMM("N907", libID=libID_MM)))
    df_MM.index = list(range(len(df_MM)))
    df = pd.concat(
        (
            df,
            df_sm_N907,
            df_sm_N926,
            df_colmar_N926_pl,
            df_colmar_N907_pl,
            # df_colmar_N926, # comment for manuscript figure
            # df_colmar_N907, # comment for manuscript figure
            df_MM,
        )
    )
    df.index = list(range(len(df)))
    return df, classes


def plot_spectrum_plots(
    lib, mixID, targetRasterSpec, pars, ar_fix, mixSubs, test_blank, show, zooms
):
    transfo = sqrt_transfo(250)
    plot_contained = True and not test_blank
    if plot_contained:
        # Show contained lib compounds
        contained = mixSubs[mixID[:-4]] if mixID[-4:] == "(dp)" else mixSubs[mixID]
        contained = [{c: s} for c, s in lib.items() if c in contained]
        for zoom in zooms:
            plotLibOnTarget(
                contained,
                targetRasterSpec,
                "metabominer",
                pars["targetSpecID"],
                densityTransform=transfo,
                assignmentRadius=ar_fix,
                # makeLabel=True, figsize=(6,4), # for legend plot
                makeLabel=False,
                zoom=zoom,
                figsize=(12, 10),
                outdir=outdir,
                plotTargetRaster=True,
                figtitle=targetRasterSpec.name + "_contained_cpds",
                show=show,
            )

    plot_picked_peaks = True and not test_blank
    if plot_picked_peaks:
        # Show provided peaklist
        mixPointSpec = loadMMMixPointSpectra()[mixID]
        peaklistTarget = {mixID: mixPointSpec}
        for zoom in zooms:
            plotLibOnTarget(
                peaklistTarget,
                targetRasterSpec,
                "metabominer",
                pars["targetSpecID"],
                densityTransform=transfo,
                makeLabel=False,
                figsize=(12, 10),
                outdir=outdir,
                zoom=zoom,
                figtitle=targetRasterSpec.name + "_pickedpeaks",
                plotTargetRaster=True,
                show=show,
            )

    plot_raster = True and not test_blank
    if plot_raster:
        use_sqrt_transfo = True
        if use_sqrt_transfo:
            # Plot raster spectrum
            def transfo_raster(x):
                res = np.zeros_like(x)
                try:
                    iter(x)
                    res[x > 0] = np.sqrt(x[x > 0])
                    res[x < 0] = -np.sqrt(-x[x < 0])
                except:
                    res = np.sqrt(x) if x > 0 else -np.sqrt(-x)
                return res

        else:
            transfo_raster = lambda x: x

        rc = np.max(np.abs(transfo_raster(targetRasterSpec.fullData)))
        # clim = (0, np.max(transfo_raster(targetRasterSpec.fullData)))
        clim = (-rc, rc)
        plot_ranges = {
            "y": (0, targetRasterSpec.FRanges[0][1]),
            "x": (0, targetRasterSpec.FRanges[1][1]),
        }
        plot_ranges = {"y": (50, 80), "x": (2.8, 5.2)}
        ax = plotRasterSpectrum(
            clim=clim,
            title=pars["targetSpecID"],
            densityTransform=transfo_raster,
            spec=targetRasterSpec,
            plotRange=plot_ranges,
            returnAx=True,
            colorbar=False,
            cmap="grey",
            show=show,
        )
        fig = ax.get_figure()
        fig.set_size_inches((12, 10))
        if use_sqrt_transfo:
            fn = outdir / (pars["targetSpecID"] + f"(sqrt-range {rc:g})" + ".png")
        else:
            fn = outdir / (pars["targetSpecID"] + f"(range {rc:g})" + ".png")
        fig.savefig(fn, dpi=600)
        print(f"Saved figure '{fn}'.\n")

        ax.legend(loc=2).remove()
        for zoom in zooms:
            ax.set_xlim(zoom["H"])
            ax.set_ylim(zoom["C"])
            fig.set_size_inches((5, 5))
            if use_sqrt_transfo:
                fn = outdir / (
                    pars["targetSpecID"]
                    + f"(sqrt-range {rc:g}, zoom {zoom['H']}x{zoom['C']})"
                    + ".png"
                )
            else:
                fn = outdir / (
                    pars["targetSpecID"]
                    + f"(range {rc:g}, zoom {zoom['H']}x{zoom['C']})"
                    + ".png"
                )
            fig.savefig(fn, dpi=600)
            print(f"Saved figure '{fn}'.\n")
        if show:
            plt.show()
        else:
            plt.close(fig)

    plot_noise = True
    if plot_noise:
        # Plot raster spectrum
        clim = (-targetRasterSpec.noiseStd, targetRasterSpec.noiseStd)
        plot_ranges = {
            "y": (0, targetRasterSpec.FRanges[0][1]),
            "x": (0, targetRasterSpec.FRanges[1][1]),
        }
        ax = plotRasterSpectrum(
            clim=clim,
            title=pars["targetSpecID"],
            densityTransform=transfo,
            spec=targetRasterSpec,
            plotRange=plot_ranges,
            returnAx=True,
            colorbar=False,
            show=show,
        )
        fig = ax.get_figure()
        fig.set_size_inches((12, 10))
        fn = outdir / (pars["targetSpecID"] + "_noise.png")
        fig.savefig(fn, dpi=600)
        print(f"Saved figure '{fn}'.\n")
        if show:
            plt.show()
        else:
            plt.close(fig)


def plot_fit_plots(
    targetPointSpec,
    targetRasterSpec,
    lib,
    mixSubs,
    pars,
    setup,
    mixID,
    test_blank,
    scanResults,
    assigned,
    arSpan,
    ar_fix,
    th_fix,
    show,
    zooms,
):

    def transfo(x):
        res = np.zeros_like(x)
        try:
            iter(x)
            res[x > 0] = np.minimum(np.sqrt(x[x > 0]), 250)
            res[x < 0] = np.maximum(-np.sqrt(-x[x < 0]), -250)
        except:
            res = (
                np.minimum(np.sqrt(x), 250) if x > 0 else np.maximum(-np.sqrt(-x), -250)
            )
        return res

    plot_scan = True
    if plot_scan:
        incrStr = "incr" if pars["incrementalFit"] else "onepass"
        isoStr = "iso" if pars["isolated_fit"] else "joint"
        setupTitle = f"{targetPointSpec.name} paramscan, {isoStr}-{incrStr}"
        axDict = plotPerformanceScanResults(
            scanResults,
            title=setupTitle,
            logAR=False,
            outdir=outdir,
            show=show,
            plotDetectionsInsteadOfPrecision=test_blank,
            verb=0,
            figsize=(5.0,2.0),
            plot_max_marker=False,
        )

    plot_slices = True
    if plot_slices:  # and plot:
        figname = outdir / (
            "Slices_"
            + pars["targetSpecID"]
            + f"({setup})"
            + "_th%g_ar%g_" % (th_fix, ar_fix)
            + ".svg"
        )
        figsize = (2.3, 2.7)
        # figsize = (3.0, 3.5)
        # figsize = (8, 7)
        plotSlices(
            scanResults,
            ar_fix=ar_fix,
            th_fix=th_fix,
            figname=figname,
            figsize=figsize,
            max_th=0.00055,
            report=False,
            show=show,
        )

    plot_assignment = True
    if plot_assignment:
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots(layout="constrained", figsize=(3, 2))
        ax.plot(arSpan * arSpan, [assigned[ar] for ar in arSpan], label="assigned")
        ax.set_xlim((min(arSpan * arSpan), max(arSpan * arSpan)))
        # ax.set_ylim((0.0, 1.0))
        ax.set_xlabel("$r^2$ (~assignment area)")
        ax.set_ylabel("total assignment")
        ax.set_title("Total assignment for " + pars["targetSpecID"])
        ax.legend()
        figname = outdir / f"ar_vs_assignment_{pars['targetSpecID']}, setup {setup}.svg"
        fig.savefig(figname)
        print(f"Saved figure '{figname}'")
        if show:
            plt.show()
        else:
            plt.close(fig)

    plot_diagnosis = True and not test_blank
    if plot_diagnosis:
        # Plot FPs, FNs, and TPs
        ID2name = buildHMDBID2NameMap(verb=0)
        scores = scanResults[ar_fix][th_fix]
        for typ in ["TP", "FP", "FN"]:
            scorelib = [{c: lib[c]} for c in scores[mixID]["classification"][typ] if c in lib]
            print(f"\n\n## {len(scores[mixID]['classification'][typ])} {typ}s of {mixID}:")
            for cd in scorelib:
                c, s = list(cd.items())[0]
                pp(
                    {
                        c
                        + " ("
                        + ID2name[c]
                        + ")": dict(weights=s.weights, coords=s.coords)
                    }
                )
                # # Debug
                # from shapely.geometry import Point
                # target = mixSpecs[mixID]
                # region = target.regions[0]
                # for pt in s.coords:
                #     print(f"Point {pt} in target region: %s"%(region.contains(Point(pt))))

            for zoom in zooms:
                plotLibOnTarget(
                    scorelib,
                    targetRasterSpec,
                    "metabominer",
                    pars["targetSpecID"],
                    assignmentRadius=ar_fix,
                    markersize=1,
                    densityTransform=transfo,
                    makeLabel=True,
                    figsize=(12, 10),
                    outdir=outdir,
                    radius_alpha=0.2,
                    zoom=zoom,
                    figtitle=targetRasterSpec.name
                    + f"_{setup}_"
                    + typ
                    + f"_ar{ar_fix:g}_th{th_fix:g}",
                    show=show,
                    plotTargetRaster=True,
                    # plot legends
                    # plotTargetRaster=False,
                    # ext="svg",
                )
        # Plot FPs vs contained:
        FPmix = makeMix(
            {c: lib[c] for c in scores[mixID]["classification"]["FP"] if c in lib}, name="FPs"
        )
        FNmix = makeMix(
            {c: lib[c] for c in scores[mixID]["classification"]["FN"] if c in lib},
            name="FNs",
        )
        TPmix = makeMix(
            {c: lib[c] for c in scores[mixID]["classification"]["TP"] if c in lib}, name="TPs"
        )
        # contained_mix = makeMix({c: lib[c] for c in mixSubs[mixID] if c in lib}, name="contained")

        for zoom in zooms:
            plotLibOnTarget(
                [{"FPs": FPmix}, {"FNs": FNmix}, {"TPs": TPmix}],
                targetRasterSpec,
                "metabominer",
                pars["targetSpecID"],
                assignmentRadius=ar_fix,
                markersize=1,
                densityTransform=transfo,
                makeLabel=True,
                figsize=(12, 10),
                outdir=outdir,
                radius_alpha=0.2,
                zoom=zoom,
                figtitle=targetRasterSpec.name
                + f"_{setup}_FPvsFNvsTP_ar{ar_fix:g}_th{th_fix:g}",
                show=show,
                plotTargetRaster=True,
                # plot legends
                # plotTargetRaster=False,
                # ext="svg",
            )


def report_performance(df, classes):
    # Report FN and FP stats
    fp_counts, fn_counts = defaultdict(list), defaultdict(list)
    for method in classes.keys():
        print(method)
        print(sorted(classes[method]["FP"]))
        for cpd in classes[method]["FP"]:
            fp_counts[cpd].append(method)
        for cpd in classes[method]["FN"]:
            fn_counts[cpd].append(method)

    ID2name = buildHMDBID2NameMap(verb=0)
    cpds = sorted(fp_counts.keys())
    counts = [len(fp_counts[cpd]) for cpd in cpds]
    ixs = np.argsort([f"{cnt}-{cpd}" for cnt,cpd in zip(counts, cpds)])
    print("\n# False Positive counts:")
    for ix in ixs:
        print(
            f"{ID2name[cpds[ix]]} (HMDB-ID: {cpds[ix]}): {counts[ix]} ({fp_counts[cpds[ix]]})"
        )
    cpds = sorted(fn_counts.keys())
    counts = [len(fn_counts[cpd]) for cpd in cpds]
    ixs = np.argsort([f"{cnt}-{cpd}" for cnt,cpd in zip(counts, cpds)])
    print("\n# False Negative counts:")
    for ix in ixs:
        print(
            f"{ID2name[cpds[ix]]} (HMDB-ID: {cpds[ix]}): {counts[ix]} ({fn_counts[cpds[ix]]})"
        )
    
    print("\nOptimal F1 scores")
    print(df.columns)
    methods = sorted(set(df["method"]))
    g = df.groupby("method").groups
    for m in methods:
        print("#", m)
        ix = g[m]
        dfg = df.loc[ix,:]
        mixIDs = sorted(set(dfg["mixID"]))
        gg = dfg.groupby("mixID").groups
        for mixID in mixIDs:
            ixx = gg[mixID]
            fscore = list(dfg.loc[ixx, "F1 score"])[0]
            th = list(dfg.loc[ixx, "th_opt"])[0]
            ar = list(dfg.loc[ixx, "ar_opt"])[0]
            print(f"{mixID}: F1 = {fscore:g} (ar={ar:g}, th={th:g})")
        print()
    
    


def run_sample_diagnosis(
    mixID="N926",
    ar_fix=0.3,
    th_fix=0.0,
    test_blank=False,
    blank_seed=None,
    plot_spectrum=True,
    plot_fit=True,
    setup=None,
    show=False,
    zooms=[],
    load=False,
    libName="metabominer",
):

    if setup == "D(dp)":
        # Use setup D on the deeppicker peaklist
        setup = "D"
        deeppicker = True
    else:
        deeppicker = False

    noiseFilterLevel = None
    smoothing = 0.0
    # smoothing = 0.5

    lib = loadMMLibrary(libName=libName)
    mixSubs = loadMMMixtureLists()

    arSpan = [
        0.001,
        0.0025,
        0.005,
        0.01,
        0.02,
        0.03,
        0.04,
        0.05,
        0.06,
        0.07,
        0.08,
        0.09,
        0.1,
        0.11,
        0.12,
        0.13,
        0.14,
        0.15,
    ]
    if mixID in ["N907", "N926"]:
        thSpan = np.linspace(0.0, 0.001, 41)
    else:
        # This has other scales
        thSpan = np.linspace(0.0, 0.015, 31)

    arSpan = np.array([np.round(ar, 7) for ar in arSpan])
    thSpan = np.array([np.round(th, 7) for th in thSpan])

    print("Check lib containment...")
    for ID, contained in zip(
        sorted(mixSubs[mixID]), np.isin(sorted(mixSubs[mixID]), list(lib))
    ):
        if not contained:
            print(f"HMDB-ID {ID} is not in MM-library.")
            if ID == "294":
                print("(HMDB-ID 294 denotes Urea)\n")

    setupSpec = case_dict[setup][mixID]
    bins = binning[mixID]
    kwargs = dict(
        assignmentRadius=None,
        incrementalFit=setupSpec["incrementalFit"],
        isolated_fit=setupSpec["isolated_fit"],
        pointTarget=setupSpec["pointTarget"],
        smoothing=smoothing,
        noiseFilterLevel=noiseFilterLevel,
        binning=bins,
        workingDir=cache_dir,
    )
    pars = prepareParameters(
        specDomain="metabominer", targetSpecID=mixID, libID=libName, **kwargs
    )

    # denoising_pars = dict(smoothRadius=1.0, noiseFilterLevel=None)
    denoising_pars = dict(smoothRadius=smoothing, noiseFilterLevel=noiseFilterLevel)
    targetRasterSpec = loadMetaboMinerSpectrum(mixID, denoising_pars=denoising_pars)

    if test_blank:
        seed = blank_seed
        blankRasterSpec = sampleBlankSpectrum(targetRasterSpec, seed)
        targetRasterSpec = blankRasterSpec
        pars["targetSpecID"] += "(blank, seed%d)" % seed

    if pars["pointTarget"]:
        if deeppicker:
            mixID += "(dp)"
            targetPointSpec = loadMMMixPointSpectra()[mixID]
        else:
            targetPointSpec = loadMMMixPointSpectra()[mixID]
        targetPointSpec.regions = []
    else:
        targetPointSpec = makeRasterPointSpectrum(
            targetRasterSpec, nbin=bins, signalThreshold=noiseFilterLevel
        )

    # Diagnostic plots prior to fit
    if plot_spectrum:
        plot_spectrum_plots(
            lib, mixID, targetRasterSpec, pars, ar_fix, mixSubs, test_blank, show, zooms
        )

    ## Run scan
    data = dict()
    assigned = {}

    if pars["pointTarget"]:
        mixSpecs = loadMMMixPointSpectra()
    else:
        pars["binning"] = binning[mixID]
        mixSpecs = collect_and_serialize_MM_grid_spectra([mixID], pars)

    if pars["incrementalFit"]:
        pars["arSpan"] = arSpan
        results = runMetaboMinerFits(
            base_pars=pars,
            lib=lib,
            mixSpecs=mixSpecs,
            mixIDs=[mixID],
            returnAllResults=True,
            load=load,
        )
        for ar in arSpan:
            pars["assignmentRadius"] = ar
            dfTargets, dfCompounds = makeResultsDF(
                [{"res": results[mixID][ar], "pars": pars}],
                originalWeights=totalWeights(lib),
            )
            data[ar] = {
                mixID: {
                    "targets": dfTargets,
                    "compounds": dfCompounds,
                    "targetSpecID": pars["targetSpecID"],
                    "targetTotalWeight": results[mixID][ar].originalWeightY,
                }
            }
    else:
        for ar in arSpan:
            pars["assignmentRadius"] = ar
            results = runMetaboMinerFits(
                # base_pars=pars, lib=lib, mixSpecs=mixSpecs, mixIDs=[mixID], load=False
                base_pars=pars,
                lib=lib,
                mixSpecs=mixSpecs,
                mixIDs=[mixID],
                load=load,
            )
            result = results[mixID]

            dfTargets, dfCompounds = makeResultsDF(
                [{"res": result, "pars": pars}], originalWeights=totalWeights(lib)
            )
            data[ar] = {
                mixID: {
                    "targets": dfTargets,
                    "compounds": dfCompounds,
                    "targetSpecID": pars["targetSpecID"],
                    "targetTotalWeight": result.originalWeightY,
                }
            }

    scanResults = defaultdict(dict)
    for ar in arSpan:
        assigned[ar] = data[ar][mixID]["targets"]["assigned"].iloc[0]
        assigned[ar] *= data[ar][mixID]["targets"]["originalWeightY"].iloc[0]
        for th in thSpan:
            scores = classificationStats(
                data[ar], pars, th, mixSubs, [mixID], scale_by_peaknr=True, verb=False
            )
            scanResults[ar][th] = scores

    if plot_fit:
        plot_fit_plots(
            targetPointSpec,
            targetRasterSpec,
            lib,
            mixSubs,
            pars,
            setup,
            mixID,
            test_blank,
            scanResults,
            assigned,
            arSpan,
            ar_fix,
            th_fix,
            show,
            zooms,
        )

    return scanResults, assigned


def selectParameters(scanResults, assigned):
    targetIDs = sorted(scanResults)
    mixIDs = sorted(set(targetID.split("_")[0] for targetID in targetIDs))
    arSpan = np.array(sorted(scanResults[targetIDs[0]]))

    slope_range = [0.12, 0.15]
    for mixID in mixIDs:
        blankID = mixID + "_blank"
        ar1, ar2 = slope_range
        # slope of ar^2 vs assined
        slope_blank = (assigned[blankID][ar2] - assigned[blankID][ar1]) / (
            ar2 * ar2 - ar1 * ar1
        )
        print(f"\n{mixID}:")
        print("   blank assignment slope over (%g, %g): %g" % (ar1, ar2, slope_blank))
        for ar1, ar2 in zip(arSpan[:-1], arSpan[1:]):
            slope_blank = (assigned[blankID][ar2] - assigned[blankID][ar1]) / (
                ar2 * ar2 - ar1 * ar1
            )
            print("       (%g, %g): %g" % (ar1, ar2, slope_blank))
        print("     mix assignment slopes:")
        for ar1, ar2 in zip(arSpan[:-1], arSpan[1:]):
            slope_mix = (assigned[mixID][ar2] - assigned[mixID][ar1]) / (
                ar2 * ar2 - ar1 * ar1
            )
            print("       (%g, %g): %g" % (ar1, ar2, slope_mix))


def buildComparisonDataframe(scanResults):
    df = {
        "libID": [],
        "F1 score": [],
        "precision": [],
        "recall": [],
        "method": [],
        "mixID": [],
        "ar_opt": [],
        "th_opt": [],
        "TP": [],
        "FN": [],
        "FP": [],
    }
    classes = defaultdict(dict)

    for setup in scanResults:
        for mixID, data in scanResults[setup].items():
            ar_opt, th_opt = None, None
            maxf1 = -np.inf
            tp, fp, fn = 0, 0, 0
            for ar in data:
                for th, scores in data[ar].items():
                    if setup[-4:] == "(dp)":
                        s = scores[mixID + "(dp)"]["classification"]
                    else:
                        s = scores[mixID]["classification"]
                    f1 = s["fscore"]
                    if f1 > maxf1:
                        maxf1 = f1
                        ar_opt, th_opt = ar, th
                        tp, fp, fn = len(s["TP"]), len(s["FP"]), len(s["FN"])
                        TPs, FPs, FNs = s["TP"], s["FP"], s["FN"]
                        recall, precision = s["recall"], s["precision"]
            df["libID"].append(metaboMinerLibNames[0])
            df["F1 score"].append(maxf1)
            df["precision"].append(precision)
            df["recall"].append(recall)
            df["TP"].append(tp)
            df["FP"].append(fp)
            df["FN"].append(fn)
            df["ar_opt"].append(ar_opt)
            df["th_opt"].append(th_opt)
            df["method"].append(setup)
            df["mixID"].append(mixID)
        classes[setup]["TP"] = TPs
        classes[setup]["FP"] = FPs
        classes[setup]["FN"] = FNs
    return pd.DataFrame(df), classes


def run(save, mixIDs, setups, ar_fix, th_fix, load, plot, show, libName=None):
    
    if libName is None:
        libName = "metabominer"

    # Zoom display for raster plots
    zooms = [{"C": [53, 80], "H": [2.8, 5.2]}, {"C": [30, 53], "H": [2.2, 3.8]}]
    scanResults, assigned = defaultdict(dict), defaultdict(dict)
    for mixID in mixIDs:
        for setup in setups:
            # # Debug: blank spectrum analysis
            # scanResults[mixID + "_blank"], assigned[mixID + "_blank"] = run_sample_diagnosis(
            #     mixID=mixID, ar_fix=ar_fix, th_fix=th_fix,
            #     test_blank=True, blank_seed=123,
            # plot_spectrum=False, plot_fit=False,
            # )
            scanResults[f"MCF-{setup}"][mixID], assigned[f"MCF-{setup}"][mixID] = (
                run_sample_diagnosis(
                    mixID=mixID,
                    ar_fix=ar_fix,
                    th_fix=th_fix,
                    setup=setup,
                    plot_spectrum=False,
                    plot_fit=plot,
                    zooms=zooms,
                    load=load,
                    show=show,
                    libName=libName
                )
            )

    df, classes = buildComparisonDataframe(scanResults)

    if save:
        tmpfn = get_mcfnmr_home() / "tmp" / f"MCFNMR_comp_df_lib_{libName}.pickle"
        with open(tmpfn, "wb") as f:
            pickle.dump((df, classes), f)
        print(f"Saved '{tmpfn}'")
    else:
        return df, classes

    # # Debug: Deprecated approach trying to detect optimal radius and threshold
    # # by looking at detections in blank spectrum. Sample specific noise
    # # characteristics are hard to emulate, though.
    # selectParameters(scanResults, assigned)


def main(load=False, plot=False, libName="MetaboMiner-all"):
    mixIDs = ["N926", "N907"]
    # mixIDs = ["N926"]
    setups = ["A", "B", "C", "D", "D(dp)"]
    # setups = ["A", "B", "C", "D"] # For manuscript figure
    # setups = ["C"] # For manuscript figure
    ar_fix = 0.05
    th_fix = 0.000125
    
    # Debug
    # save==True generates file tmp/comp_df.pickle,
    # used to avoid lengthy collection of scan results for MCF
    save = True
    show = False
    res = run(
        save,
        mixIDs=mixIDs,
        setups=setups,
        ar_fix=ar_fix,
        th_fix=th_fix,
        load=load,
        plot=plot,
        show=show,
        libName=libName,
    )

    if not save:
        df, classes_mcf = res
    else:
        # Expects saved result from run(True)
        tmpfn = get_mcfnmr_home() / "tmp" / f"MCFNMR_comp_df_lib_{libName}.pickle"
        with open(tmpfn, "rb") as f:
            df, classes_mcf = pickle.load(f)
            print(f"Loaded '{tmpfn}'")
    
    colmar_match_cutoffs = {"N926(dp)": 0.9, "N907(dp)": 0.9,
                            "N926": 0.8, "N907": 0.8}
    libID_MM = "MetaboMiner-all" if libName.lower() == "metabominer" else libName
    df, classes = mergeOtherResults(df, colmar_match_cutoffs=colmar_match_cutoffs,
                                    libID_MM = libID_MM)
    classes.update(classes_mcf)

    import seaborn as sns


    plot_comparison = True
    if plot_comparison:
        pal = sns.color_palette("magma", n_colors=7)
        pal = pal[:3] + [pal[3]] * 2 + pal[4:6] + pal[5:] # comment for manuscript figure
        plotMetaboMinerComparison(
            None,
            df,
            plotOnlyFScore=True,
            title=f"Comparison plasma sample (lib {libName})",
            show=show,
            show_legend=False,
            # show_legend=True, # comment for manuscript figure
            plot_width=3.7,
            plot_height=2.3,
            pal=pal,
        )

    report = True
    if report:
        report_performance(df, classes)


if __name__ == "__main__":
    # libName = "metabominer" # → "MetaboMiner-all" (version for paper)
    # libName = "MetaboMiner-common"
    # libName = "MetaboMiner-plasma-all"
    # libName = "MetaboMiner-plasma-common"

    # main(load=True, plot=False, libName="metabominer")
    main(load=True, plot=False, libName="MetaboMiner-plasma-common")



import os
import numpy as np
from mcfnmr.utils.system import get_mcfnmr_home

# Global debug-level in {0,1,2} controls verbosity
# and runtime sanity checks
DEBUG = 1

if DEBUG:
    print(
        "config: Enforcing numpy legacy print mode to ensure that numpy floats are printed without type information."
    )
np.set_printoptions(legacy="1.21")

# mcfNMR repository url
REPO_URL = "https://github.com/GeoMetabolomics-ICBM/mcfNMR"

MCFNMR_HOME = get_mcfnmr_home()
OUTDIR = MCFNMR_HOME / "output"
TEMPDIR = MCFNMR_HOME / "tmp"
DATADIR = MCFNMR_HOME / "data"

OLDBDATADIR = DATADIR / "OLDBdata"
OLDBPEAKSDIR = OLDBDATADIR / "peaks"
OLDB1DTESTDIR = OLDBDATADIR / "test1D"
OLDBFULLSPECTRADIR = OLDBDATADIR / "fullspectra"

METABOMINER_DIR = DATADIR / "MetaboMiner"
UCSFDIR = METABOMINER_DIR / "ucsf"
PROCESSED_UCSFDIR = METABOMINER_DIR / "processedUCSF"
SPARKYDIR = METABOMINER_DIR / "sparky"
MM_LIBDIR = METABOMINER_DIR / "metabominer" / "lib"
MM_MIXLISTSDIR = METABOMINER_DIR / "mixture_contents"
MM_PEAKLISTDIR = METABOMINER_DIR / "examples" / "peaklist"
UCSFDATA_BIN = SPARKYDIR / "bin" / "ucsfdata"

SMARTMINER_DATADIR = DATADIR / "SMART-Miner"

COLMAR_DATADIR = DATADIR / "COLMAR"

# Maximal absorption cost (assumed larger than any possibly encountered point distance)
MAX_ABSORPTON_COST = 1e6

# Default scales for HSQC spectra
DEFAULT_SCALE_C = 10.0
DEFAULT_SCALE_H = 1.0

if not OUTDIR.exists():
    os.makedirs(OUTDIR)
    print(f"Created directory '{OUTDIR}'")
if not TEMPDIR.exists():
    os.makedirs(TEMPDIR)
    print(f"Created directory '{TEMPDIR}'")

# This defines default parameters for MCF recombinations
default_pars = dict(
    assignmentRadius=0.01,
    absorptionCost=None,
    regionDefs=None,
    gridInfoY=None,
    # Preprocessing step for full raster spectrum
    smoothing=0,
    # peak resolution of raster to point conversion
    binning=(2, 2),
    # Whether to use the peak list dat of the target (or the raster data)
    # (Only effective if both are available)
    pointTarget=True,
    # Whether to do a simulataneous isolated fit for all compounds
    isolated_fit=False,
    # threshold (relative to noiseStd) for cutting off raster spectrum signal
    noiseFilterLevel=None,
    distPars={"scalex": DEFAULT_SCALE_C, "scaley": DEFAULT_SCALE_H},
)


# Column name candidates for the different dimensions
# (case-insensitive
csv_column_name_links = {
    "weight": ["w", "weights", "weight", "integral", "i", "intensity"],
    "1H": ["1h", "ppm h"],
    "13C": ["13c", "ppm c"],
}
